# -*- coding: utf-8 -*-

from typing import Union, Callable, Dict, Sequence

from brainpy import errors, math as bm
from brainpy.base import Collector
from brainpy.integrators import constants, utils, joint_eq
from brainpy.integrators.sde.base import SDEIntegrator
from .generic import register_sde_integrator

__all__ = [
  'Euler',
  'Heun',
  'Milstein',
  'ExponentialEuler',
]


def df_and_dg(code_lines, variables, parameters):
  # 1. df
  # df = f(x, t, *args)
  all_df = [f'{var}_df' for var in variables]
  code_lines.append(f'  {", ".join(all_df)} = f({", ".join(variables + parameters)})')

  # 2. dg
  # dg = g(x, t, *args)
  all_dg = [f'{var}_dg' for var in variables]
  code_lines.append(f'  {", ".join(all_dg)} = g({", ".join(variables + parameters)})')
  code_lines.append('  ')


def dfdt(code_lines, variables):
  for var in variables:
    code_lines.append(f'  {var}_dfdt = {var}_df * {constants.DT}')
  code_lines.append('  ')


def noise_terms(code_lines, variables):
  # num_vars = len(variables)
  # if num_vars > 1:
  #     code_lines.append(f'  all_dW = math.normal(0.0, dt_sqrt, ({num_vars},)+math.shape({variables[0]}_dg))')
  #     for i, var in enumerate(variables):
  #         code_lines.append(f'  {var}_dW = all_dW[{i}]')
  # else:
  #     var = variables[0]
  #     code_lines.append(f'  {var}_dW = math.normal(0.0, dt_sqrt, math.shape({var}))')
  # code_lines.append('  ')

  for var in variables:
    code_lines.append(f'  {var}_dW = random.normal(0.000, dt_sqrt, math.shape({var})).value')
  code_lines.append('  ')


class Euler(SDEIntegrator):
  def __init__(self, f, g, dt=None, name=None, show_code=False,
               var_type=None, intg_type=None, wiener_type=None,
               state_delays=None):
    super(Euler, self).__init__(f=f, g=g, dt=dt, show_code=show_code, name=name,
                                var_type=var_type, intg_type=intg_type,
                                wiener_type=wiener_type, state_delays=state_delays)
    self.build()

  def build(self):
    self.code_lines.append(f'  {constants.DT}_sqrt = {constants.DT} ** 0.5')

    # 2.1 df, dg
    df_and_dg(self.code_lines, self.variables, self.parameters)

    # 2.2 dfdt
    dfdt(self.code_lines, self.variables)

    # 2.3 dW
    noise_terms(self.code_lines, self.variables)

    # 2.3 dgdW
    # ----
    # SCALAR_WIENER : dg * dW
    # VECTOR_WIENER : math.sum(dg * dW, axis=-1)

    if self.wiener_type == constants.SCALAR_WIENER:
      for var in self.variables:
        self.code_lines.append(f'  {var}_dgdW = {var}_dg * {var}_dW')
    else:
      for var in self.variables:
        self.code_lines.append(f'  {var}_dgdW = math.sum({var}_dg * {var}_dW, axis=-1)')
    self.code_lines.append('  ')

    if self.intg_type == constants.ITO_SDE:
      # 2.4 new var
      # ----
      # y = x + dfdt + dgdW
      for var in self.variables:
        self.code_lines.append(f'  {var}_new = {var} + {var}_dfdt + {var}_dgdW')
      self.code_lines.append('  ')

    elif self.intg_type == constants.STRA_SDE:
      # 2.4  y_bar = x + math.sum(dgdW, axis=-1)
      all_bar = [f'{var}_bar' for var in self.variables]
      for var in self.variables:
        self.code_lines.append(f'  {var}_bar = {var} + {var}_dgdW')
      self.code_lines.append('  ')

      # 2.5  dg_bar = g(y_bar, t, *args)
      all_dg_bar = [f'{var}_dg_bar' for var in self.variables]
      self.code_lines.append(f'  {", ".join(all_dg_bar)} = g({", ".join(all_bar + self.parameters)})')

      # 2.6 dgdW2
      # ----
      # SCALAR_WIENER : dgdW2 = dg_bar * dW
      # VECTOR_WIENER : dgdW2 = math.sum(dg_bar * dW, axis=-1)
      if self.wiener_type == constants.SCALAR_WIENER:
        for var in self.variables:
          self.code_lines.append(f'  {var}_dgdW2 = {var}_dg_bar * {var}_dW')
      else:
        for var in self.variables:
          self.code_lines.append(f'  {var}_dgdW2 = math.sum({var}_dg_bar * {var}_dW, axis=-1)')
      self.code_lines.append('  ')

      # 2.7 new var
      # ----
      # y = x + dfdt + 0.5 * (dgdW + dgdW2)
      for var in self.variables:
        self.code_lines.append(f'  {var}_new = {var} + {var}_dfdt + 0.5 * ({var}_dgdW + {var}_dgdW2)')
      self.code_lines.append('  ')
    else:
      raise ValueError(f'Unknown SDE_INT type: {self.intg_type}. We only '
                       f'supports {constants.SUPPORTED_INTG_TYPE}.')

    # returns
    new_vars = [f'{var}_new' for var in self.variables]
    self.code_lines.append(f'  return {", ".join(new_vars)}')

    # return and compile
    self.integral = utils.compile_code(
      code_scope={k: v for k, v in self.code_scope.items()},
      code_lines=self.code_lines,
      show_code=self.show_code,
      func_name=self.func_name)


register_sde_integrator('euler', Euler)


class Heun(Euler):
  def __init__(self, f, g, dt=None, name=None, show_code=False,
               var_type=None, intg_type=None, wiener_type=None,
               state_delays=None):
    if intg_type != constants.STRA_SDE:
      raise errors.IntegratorError(f'Heun method only supports Stranovich integral of SDEs, '
                                   f'but we got {intg_type} integral.')
    super(Heun, self).__init__(f=f, g=g, dt=dt, show_code=show_code, name=name,
                               var_type=var_type, intg_type=intg_type,
                               wiener_type=wiener_type, state_delays=state_delays)
    self.build()


register_sde_integrator('heun', Heun)


class Milstein(SDEIntegrator):
  def __init__(self, f, g, dt=None, name=None, show_code=False,
               var_type=None, intg_type=None, wiener_type=None,
               state_delays=None):
    super(Milstein, self).__init__(f=f, g=g, dt=dt, show_code=show_code, name=name,
                                   var_type=var_type, intg_type=intg_type,
                                   wiener_type=wiener_type, state_delays=state_delays)
    self.build()

  def build(self):
    # 2. code lines
    self.code_lines.append(f'  {constants.DT}_sqrt = {constants.DT} ** 0.5')

    # 2.1 df, dg
    df_and_dg(self.code_lines, self.variables, self.parameters)

    # 2.2 dfdt
    dfdt(self.code_lines, self.variables)

    # 2.3 dW
    noise_terms(self.code_lines, self.variables)

    # 2.3 dgdW
    # ----
    # dg * dW
    for var in self.variables:
      self.code_lines.append(f'  {var}_dgdW = {var}_dg * {var}_dW')
    self.code_lines.append('  ')

    # 2.4  df_bar = x + dfdt + math.sum(dg * dt_sqrt, axis=-1)
    all_df_bar = [f'{var}_df_bar' for var in self.variables]
    if self.wiener_type == constants.SCALAR_WIENER:
      for var in self.variables:
        self.code_lines.append(f'  {var}_df_bar = {var} + {var}_dfdt + {var}_dg * {constants.DT}_sqrt')
    else:
      for var in self.variables:
        self.code_lines.append(f'  {var}_df_bar = {var} + {var}_dfdt + math.sum('
                               f'{var}_dg * {constants.DT}_sqrt, axis=-1)')

    # 2.5  dg_bar = g(y_bar, t, *args)
    all_dg_bar = [f'{var}_dg_bar' for var in self.variables]
    self.code_lines.append(f'  {", ".join(all_dg_bar)} = g({", ".join(all_df_bar + self.parameters)})')
    self.code_lines.append('  ')

    # 2.6 dgdW2
    # ----
    # dgdW2 = 0.5 * (dg_bar - dg) * (dW * dW / dt_sqrt - dt_sqrt)
    if self.intg_type == constants.ITO_SDE:
      for var in self.variables:
        self.code_lines.append(f'  {var}_dgdW2 = 0.5 * ({var}_dg_bar - {var}_dg) * '
                               f'({var}_dW * {var}_dW / {constants.DT}_sqrt - {constants.DT}_sqrt)')
    elif self.intg_type == constants.STRA_SDE:
      for var in self.variables:
        self.code_lines.append(f'  {var}_dgdW2 = 0.5 * ({var}_dg_bar - {var}_dg) * '
                               f'{var}_dW * {var}_dW / {constants.DT}_sqrt')
    else:
      raise ValueError(f'Unknown SDE_INT type: {self.intg_type}')
    self.code_lines.append('  ')

    # 2.7 new var
    # ----
    # SCALAR_WIENER : y = x + dfdt + dgdW + dgdW2
    # VECTOR_WIENER : y = x + dfdt + math.sum(dgdW + dgdW2, axis=-1)
    if self.wiener_type == constants.SCALAR_WIENER:
      for var in self.variables:
        self.code_lines.append(f'  {var}_new = {var} + {var}_dfdt + {var}_dgdW + {var}_dgdW2')
    elif self.wiener_type == constants.VECTOR_WIENER:
      for var in self.variables:
        self.code_lines.append(f'  {var}_new = {var} + {var}_dfdt + math.sum({var}_dgdW + {var}_dgdW2, axis=-1)')
    else:
      raise ValueError(f'Unknown Wiener Process : {self.wiener_type}')
    self.code_lines.append('  ')

    # returns
    new_vars = [f'{var}_new' for var in self.variables]
    self.code_lines.append(f'  return {", ".join(new_vars)}')

    # return and compile
    self.integral = utils.compile_code(
      code_scope={k: v for k, v in self.code_scope.items()},
      code_lines=self.code_lines,
      show_code=self.show_code,
      func_name=self.func_name)


register_sde_integrator('milstein', Milstein)


class ExponentialEuler(SDEIntegrator):
  r"""First order, explicit exponential Euler method.

  For a SDE equation of the form

  .. math::

      d y=(Ay+ F(y))dt + g(y)dW(t) = f(y)dt + g(y)dW(t), \quad y(0)=y_{0}

  its schema is given by [1]_

  .. math::

      y_{n+1} & =e^{\Delta t A}(y_{n}+ g(y_n)\Delta W_{n})+\varphi(\Delta t A) F(y_{n}) \Delta t \\
       &= y_n + \Delta t \varphi(\Delta t A) f(y) + e^{\Delta t A}g(y_n)\Delta W_{n}

  where :math:`\varphi(z)=\frac{e^{z}-1}{z}`.

  References
  ----------
  .. [1] Erdoğan, Utku, and Gabriel J. Lord. "A new class of exponential integrators for stochastic
         differential equations with multiplicative noise." arXiv preprint arXiv:1608.07096 (2016).
  """

  def __init__(
      self,
      f: Callable,
      g: Callable,
      dt: float = None,
      name: str = None,
      show_code: bool = False,
      var_type: str = None,
      intg_type: str = None,
      wiener_type: str = None,
      dyn_vars: Union[bm.Variable, Sequence[bm.Variable], Dict[str, bm.Variable]] = None,
      state_delays: Dict[str, bm.AbstractDelay] = None
  ):
    super(ExponentialEuler, self).__init__(f=f,
                                           g=g,
                                           dt=dt,
                                           show_code=show_code,
                                           name=name,
                                           var_type=var_type,
                                           intg_type=intg_type,
                                           wiener_type=wiener_type,
                                           state_delays=state_delays)

    if self.intg_type == constants.STRA_SDE:
      raise NotImplementedError(f'{self.__class__.__name__} does not support integral type of {constants.STRA_SDE}. '
                                f'It only supports {constants.ITO_SDE} now. ')
    self.dyn_vars = dyn_vars

    # build the integrator
    self.code_lines = []
    self.code_scope = {}
    self.integral = self.build()

  def build(self):
    all_vars, all_pars = [], []
    integrals, arg_names = [], []
    a = self._build_integrator(self.f)
    for integral, vars, _ in a:
      integrals.append(integral)
      for var in vars:
        if var not in all_vars:
          all_vars.append(var)
    for _, vars, pars in a:
      for par in pars:
        if (par not in all_vars) and (par not in all_pars):
          all_pars.append(par)
      arg_names.append(vars + pars + ['dt'])
    all_pars.append('dt')
    all_vps = all_vars + all_pars

    def integral_func(*args, **kwargs):
      # format arguments
      params_in = Collector()
      for i, arg in enumerate(args):
        params_in[all_vps[i]] = arg
      params_in.update(kwargs)
      dt = params_in.pop('dt', self.dt)

      # diffusion part
      noises = self.g(**params_in)

      # call integrals
      results = []
      params_in['dt'] = dt
      for i, int_fun in enumerate(integrals):
        _key = arg_names[i][0]
        r = int_fun(params_in[_key], **{arg: params_in[arg] for arg in arg_names[i][1:] if arg in params_in})
        if self.wiener_type == constants.SCALAR_WIENER:
          n = noises[i]
        else:
          if bm.ndim(noises[i]) != bm.ndim(r) + 1:
            raise ValueError(f'The dimension of the noise does not match when setting {constants.VECTOR_WIENER}. '
                             f'We got the dimension of noise {bm.ndim(noises[i])}, but we expect {bm.ndim(r) + 1}.')
          n = bm.sum(noises[i], axis=0)
        n = n * self.rng.randn(*bm.shape(r)) * bm.sqrt(params_in['dt'])
        results.append(r + n)
      return results if isinstance(self.f, joint_eq.JointEq) else results[0]

    return integral_func

  def _build_integrator(self, f):
    if isinstance(f, joint_eq.JointEq):
      results = []
      for sub_eq in f.eqs:
        results.extend(self._build_integrator(sub_eq))
      return results

    else:
      vars, pars, _ = utils.get_args(f)

      # checking
      if len(vars) != 1:
        raise errors.DiffEqError(constants.exp_error_msg.format(cls=self.__class__.__name__,
                                                                vars=str(vars),
                                                                eq=str(f)))

      # gradient function
      value_and_grad = bm.vector_grad(f, argnums=0, dyn_vars=self.dyn_vars, return_value=True)

      # integration function
      def integral(*args, **kwargs):
        assert len(args) > 0
        dt = kwargs.pop('dt', self.dt)
        linear, derivative = value_and_grad(*args, **kwargs)
        phi = bm.where(linear == 0., bm.ones_like(linear), (bm.exp(dt * linear) - 1) / (dt * linear))
        return args[0] + dt * phi * derivative

      return [(integral, vars, pars), ]


register_sde_integrator('exponential_euler', ExponentialEuler)
register_sde_integrator('exp_euler', ExponentialEuler)
register_sde_integrator('exp_euler_auto', ExponentialEuler)
register_sde_integrator('exp_auto', ExponentialEuler)



# -*- coding: utf-8 -*-

"""
This module provides numerical methods for integrating Caputo fractional derivative equations.

"""

from typing import Union, Dict

import jax.numpy as jnp

import brainpy.math as bm
from brainpy import check
from brainpy.errors import UnsupportedError
from brainpy.integrators.constants import DT
from brainpy.integrators.utils import check_inits, format_args
from brainpy.tools.errors import check_error_in_jit
from .base import FDEIntegrator
from .generic import register_fde_integrator

__all__ = [
  'CaputoEuler',
  'CaputoL1Schema',
]


class CaputoEuler(FDEIntegrator):
  r"""One-step Euler method for Caputo fractional differential equations.

  Given a fractional initial value problem,

  .. math::

     D_{*}^{\alpha} y(t)=f(t, y(t)), \quad y^{(k)}(0)=y_{0}^{(k)}, \quad k=0,1, \ldots,\lceil\alpha\rceil-1

  where the :math:`y_0^{(k)}` ay be arbitrary real numbers and where :math:`\alpha>0`.
  :math:`D_{*}^{\alpha}` denotes the differential operator in the sense of Caputo, defined
  by

  .. math::

     D_{*}^{\alpha} z(t)=J^{n-\alpha} D^{n} z(t)

  where :math:`n:=\lceil\alpha\rceil` is the smallest integer :math:`\geqslant \alpha`,
  Here :math:`D^n` is the usual differential operator of (integer) order :math:`n`,
  and for :math:`\mu > 0`, :math:`J^{\mu}` is the Riemann–Liouville integral operator
  of order :math:`\mu`, defined by

  .. math::

     J^{\mu} z(t)=\frac{1}{\Gamma(\mu)} \int_{0}^{t}(t-u)^{\mu-1} z(u) \mathrm{d} u

  The one-step Euler method for fractional differential equation is defined as

  .. math::

     y_{k+1} = y_0 + \frac{1}{\Gamma(\alpha)} \sum_{j=0}^{k} b_{j, k+1} f\left(t_{j}, y_{j}\right).

  where

  .. math::

     b_{j, k+1}=\frac{h^{\alpha}}{\alpha}\left((k+1-j)^{\alpha}-(k-j)^{\alpha}\right).


  Examples
  --------

  >>> import brainpy as bp
  >>>
  >>> a, b, c = 10, 28, 8 / 3
  >>> def lorenz(x, y, z, t):
  >>>   dx = a * (y - x)
  >>>   dy = x * (b - z) - y
  >>>   dz = x * y - c * z
  >>>   return dx, dy, dz
  >>>
  >>> duration = 30.
  >>> dt = 0.005
  >>> inits = [1., 0., 1.]
  >>> f = bp.fde.CaputoEuler(lorenz, alpha=0.97, num_step=int(duration / dt), inits=inits)
  >>> runner = bp.integrators.IntegratorRunner(f, monitors=list('xyz'), dt=dt, inits=inits)
  >>> runner.run(duration)
  >>>
  >>> import matplotlib.pyplot as plt
  >>> plt.plot(runner.mon.x.flatten(), runner.mon.z.flatten())
  >>> plt.show()


  Parameters
  ----------
  f : callable
    The derivative function.
  alpha: int, float, jnp.ndarray, bm.ndarray, sequence
    The fractional-order of the derivative function. Should be in the range of ``(0., 1.)``.
  num_step: int
    The total time step of the simulation.
  inits: sequence
    A sequence of the initial values for variables.
  dt: float, int
    The numerical precision.
  name: str
    The integrator name.

  References
  ----------
  .. [1] Li, Changpin, and Fanhai Zeng. "The finite difference methods for fractional
         ordinary differential equations." Numerical Functional Analysis and
         Optimization 34.2 (2013): 149-179.
  .. [2] Diethelm, Kai, Neville J. Ford, and Alan D. Freed. "Detailed error analysis
         for a fractional Adams method." Numerical algorithms 36.1 (2004): 31-52.
  """

  def __init__(
      self,
      f,
      alpha,
      num_step,
      inits,
      dt=None,
      name=None,
      state_delays: Dict[str, Union[bm.LengthDelay, bm.TimeDelay]] = None,
  ):
    super(CaputoEuler, self).__init__(f=f,
                                      alpha=alpha,
                                      dt=dt,
                                      name=name,
                                      num_step=num_step,
                                      state_delays=state_delays)

    # fractional order
    if not bm.all(bm.logical_and(self.alpha < 1, self.alpha > 0)):
      raise UnsupportedError(f'Only support the fractional order in (0, 1), '
                             f'but we got {self.alpha}.')

    # initial values
    self.inits = check_inits(inits, self.variables)

    # coefficients
    from scipy.special import rgamma
    rgamma_alpha = bm.asarray(rgamma(bm.as_numpy(self.alpha)))
    ranges = bm.asarray([bm.arange(num_step + 1) for _ in self.variables]).T
    coef = rgamma_alpha * bm.diff(bm.power(ranges, self.alpha), axis=0)
    self.coef = bm.flip(coef, axis=0)

    # variable states
    self.f_states = {v: bm.Variable(bm.zeros((num_step,) + self.inits[v].shape))
                     for v in self.variables}
    self.register_implicit_vars(self.f_states)
    self.idx = bm.Variable(bm.asarray([1]))

    self.set_integral(self._integral_func)

  def _check_step(self, args):
    dt, t = args
    raise ValueError(f'The maximum number of step is {self.num_step}, '
                     f'however, the current time {t} require a time '
                     f'step number {t / dt}.')

  def _integral_func(self, *args, **kwargs):
    # format arguments
    all_args = format_args(args, kwargs, self.arg_names)
    t = all_args['t']
    dt = all_args.pop(DT, self.dt)
    if check.is_checking():
      check_error_in_jit(self.num_step * dt < t, self._check_step, (dt, t))

    # derivative values
    devs = self.f(**all_args)
    if len(self.variables) == 1:
      if not isinstance(devs, (bm.ndarray, jnp.ndarray)):
        raise ValueError('Derivative values must be a tensor when there '
                         'is only one variable in the equation.')
      devs = {self.variables[0]: devs}
    else:
      if not isinstance(devs, (tuple, list)):
        raise ValueError('Derivative values must be a list/tuple of tensors '
                         'when there are multiple variables in the equation.')
      devs = {var: devs[i] for i, var in enumerate(self.variables)}

    # function states
    for key in self.variables:
      self.f_states[key][self.idx[0]] = devs[key]

    # integral results
    integrals = []
    idx = ((self.num_step - 1 - self.idx) + bm.arange(self.num_step)) % self.num_step
    for i, key in enumerate(self.variables):
      integral = self.inits[key] + self.coef[idx, i] @ self.f_states[key]
      integrals.append(integral * (dt ** self.alpha[i] / self.alpha[i]))
    self.idx.value = (self.idx + 1) % self.num_step

    # return integrals
    if len(self.variables) == 1:
      return integrals[0]
    else:
      return integrals


register_fde_integrator(name='CaputoEuler', integrator=CaputoEuler)


class CaputoL1Schema(FDEIntegrator):
  r"""The L1 scheme method for the numerical approximation of the Caputo
  fractional-order derivative equations [3]_.

  For the fractional order :math:`0<\alpha<1`, let the fractional derivative of variable
  :math:`x(t)` be

  .. math::

     \frac{d^{\alpha} x}{d t^{\alpha}}=F(x, t)

  The Caputo definition of the fractional derivative for variable :math:`x` is

  .. math::

     \frac{d^{\alpha} x}{d t^{\alpha}}=\frac{1}{\Gamma(1-\alpha)} \int_{0}^{t} \frac{x^{\prime}(u)}{(t-u)^{\alpha}} d u

  where :math:`\Gamma` is the Gamma function.

  The fractional-order derivative is capable of integrating the activity of the
  function over all past activities weighted by a function that follows a power-law.
  Using one of the numerical methods, the L1 scheme method [3]_, the numerical
  approximation of the fractional-order derivative of :math:`x` is

  .. math::

     \frac{d^{\alpha} \chi}{d t^{\alpha}} \approx \frac{(d t)^{-\alpha}}{\Gamma(2-\alpha)}\left[\sum_{k=0}^{N-1}\left[x\left(t_{k+1}\right)-
     \mathrm{x}\left(t_{k}\right)\right]\left[(N-k)^{1-\alpha}-(N-1-k)^{1-\alpha}\right]\right]

  Therefore, the numerical solution of original system is given by

  .. math::

     x\left(t_{N}\right) \approx d t^{\alpha} \Gamma(2-\alpha) F(x, t)+x\left(t_{N-1}\right)-
     \left[\sum_{k=0}^{N-2}\left[x\left(t_{k+1}\right)-x\left(t_{k}\right)\right]\left[(N-k)^{1-\alpha}-(N-1-k)^{1-\alpha}\right]\right]

  Hence, the solution of the fractional-order derivative can be described as the
  difference between the *Markov term* and the *memory trace*. The *Markov term*
  weighted by the gamma function is

  .. math::

     \text { Markov term }=d t^{\alpha} \Gamma(2-\alpha) F(x, t)+x\left(t_{N-1}\right)

  The memory trace (:math:`x`-memory trace since it is related to variable :math:`x`) is

  .. math::

     \text { Memory trace }=\sum_{k=0}^{N-2}\left[x\left(t_{k+1}\right)-x\left(t_{k}\right)\right]\left[(N-k)^{1-\alpha}-(N-(k+1))^{1-\alpha}\right]

  The memory trace integrates all the past activity and captures the long-term
  history of the system. For :math:`\alpha=1`, the memory trace is 0 for any
  time :math:`t`. When the fractional order :math:`\alpha` is decreased from 1,
  the memory trace non-linearly increases from 0, and its dynamics strongly
  depends on time. Thus, the fractional order dynamics strongly deviates
  from the first order dynamics.


  Examples
  --------

  >>> import brainpy as bp
  >>>
  >>> a, b, c = 10, 28, 8 / 3
  >>> def lorenz(x, y, z, t):
  >>>   dx = a * (y - x)
  >>>   dy = x * (b - z) - y
  >>>   dz = x * y - c * z
  >>>   return dx, dy, dz
  >>>
  >>> duration = 30.
  >>> dt = 0.005
  >>> inits = [1., 0., 1.]
  >>> f = bp.fde.CaputoL1Schema(lorenz, alpha=0.99, num_step=int(duration / dt), inits=inits)
  >>> runner = bp.integrators.IntegratorRunner(f, monitors=list('xz'), dt=dt, inits=inits)
  >>> runner.run(duration)
  >>>
  >>> import matplotlib.pyplot as plt
  >>> plt.plot(runner.mon.x.flatten(), runner.mon.z.flatten())
  >>> plt.show()


  Parameters
  ----------
  f : callable
    The derivative function.
  alpha: int, float, jnp.ndarray, bm.ndarray, sequence
    The fractional-order of the derivative function. Should be in the range of ``(0., 1.]``.
  num_step: int
    The total time step of the simulation.
  inits: sequence
    A sequence of the initial values for variables.
  dt: float, int
    The numerical precision.
  name: str
    The integrator name.

  References
  ----------
  .. [3] Oldham, K., & Spanier, J. (1974). The fractional calculus theory
         and applications of differentiation and integration to arbitrary
         order. Elsevier.
  """

  def __init__(
      self,
      f,
      alpha,
      num_step,
      inits,
      dt=None,
      name=None,
      state_delays: Dict[str, Union[bm.LengthDelay, bm.TimeDelay]] = None,
  ):
    super(CaputoL1Schema, self).__init__(f=f,
                                         alpha=alpha,
                                         dt=dt,
                                         name=name,
                                         num_step=num_step,
                                         state_delays=state_delays)

    # fractional order
    if not bm.all(bm.logical_and(self.alpha <= 1, self.alpha > 0)):
      raise UnsupportedError(f'Only support the fractional order in (0, 1), '
                             f'but we got {self.alpha}.')
    from scipy.special import gamma
    self.gamma_alpha = bm.asarray(gamma(bm.as_numpy(2 - self.alpha)))

    # initial values
    inits = check_inits(inits, self.variables)
    self.inits = {v: bm.Variable(inits[v]) for v in self.variables}
    self.register_implicit_vars(self.inits)

    # coefficients
    ranges = bm.asarray([bm.arange(1, num_step + 2) for _ in self.variables]).T
    coef = bm.diff(bm.power(ranges, 1 - self.alpha), axis=0)
    self.coef = bm.flip(coef, axis=0)

    # variable states
    self.diff_states = {v + "_diff": bm.Variable(bm.zeros((num_step,) + self.inits[v].shape,
                                                          dtype=self.inits[v].dtype))
                        for v in self.variables}
    self.register_implicit_vars(self.diff_states)
    self.idx = bm.Variable(bm.asarray([self.num_step - 1]))

    # integral function
    self.set_integral(self._integral_func)

  def reset(self, inits):
    """Reset function."""
    self.idx.value = bm.asarray([self.num_step - 1])
    inits = check_inits(inits, self.variables)
    for key, value in inits.items():
      self.inits[key].value = value
    for key, val in inits.items():
      self.diff_states[key + "_diff"].value = bm.zeros((self.num_step,) + val.shape, dtype=val.dtype)

  def hists(self, var=None, numpy=True):
    """Get the recorded history values."""
    if var is None:
      hists_ = {k: bm.vstack([self.inits[k], self.diff_states[k + '_diff']])
                for k in self.variables}
      hists_ = {k: bm.cumsum(v, axis=0) for k, v in hists_.items()}
      if numpy:
        hists_ = {k: v.numpy() for k, v in hists_}
      return hists_
    else:
      assert var in self.variables, (f'"{var}" is not defined in equation '
                                     f'variables: {self.variables}')
      hists_ = bm.vstack([self.inits[var], self.diff_states[var + '_diff']])
      hists_ = bm.cumsum(hists_, axis=0)
      if numpy:
        hists_ = hists_.numpy()
      return hists_

  def _check_step(self, args):
    dt, t = args
    raise ValueError(f'The maximum number of step is {self.num_step}, '
                     f'however, the current time {t} require a time '
                     f'step number {t / dt}.')

  def _integral_func(self, *args, **kwargs):
    # format arguments
    all_args = format_args(args, kwargs, self.arg_names)
    t = all_args['t']
    dt = all_args.pop(DT, self.dt)
    if check.is_checking():
      check_error_in_jit(self.num_step * dt < t, self._check_step, (dt, t))

    # derivative values
    devs = self.f(**all_args)
    if len(self.variables) == 1:
      if not isinstance(devs, (bm.ndarray, jnp.ndarray)):
        raise ValueError('Derivative values must be a tensor when there '
                         'is only one variable in the equation.')
      devs = {self.variables[0]: devs}
    else:
      if not isinstance(devs, (tuple, list)):
        raise ValueError('Derivative values must be a list/tuple of tensors '
                         'when there are multiple variables in the equation.')
      devs = {var: devs[i] for i, var in enumerate(self.variables)}

    # integral results
    integrals = []
    idx = ((self.num_step - 1 - self.idx) + bm.arange(self.num_step)) % self.num_step
    for i, key in enumerate(self.variables):
      self.diff_states[key + '_diff'][self.idx[0]] = all_args[key] - self.inits[key]
      self.inits[key].value = all_args[key]
      markov_term = dt ** self.alpha[i] * self.gamma_alpha[i] * devs[key] + all_args[key]
      memory_trace = self.coef[idx, i] @ self.diff_states[key + '_diff']
      integral = markov_term - memory_trace
      integrals.append(integral)
    self.idx.value = (self.idx + 1) % self.num_step

    # return integrals
    if len(self.variables) == 1:
      return integrals[0]
    else:
      return integrals


register_fde_integrator(name='CaputoL1', integrator=CaputoL1Schema)
register_fde_integrator(name='CaputoL1Schema', integrator=CaputoL1Schema)



# -*- coding: utf-8 -*-

from typing import Union, Callable

import brainpy.math as bm
from brainpy.dyn.base import ConNeuGroup
from brainpy.initialize import Initializer, init_param
from brainpy.integrators import odeint
from brainpy.types import Shape, Tensor
from .base import IonChannel

__all__ = [
  'IhChannel',
  'Ih',
]


class IhChannel(IonChannel):
  """Base class for Ih channel models."""
  master_cls = ConNeuGroup


class Ih(IhChannel):
  r"""The hyperpolarization-activated cation current model.

  The hyperpolarization-activated cation current model is adopted from (Huguenard, et, al., 1992) [1]_.
  Its dynamics is given by:

  .. math::

      \begin{aligned}
      I_h &= g_{\mathrm{max}} p
      \\
      \frac{dp}{dt} &= \phi \frac{p_{\infty} - p}{\tau_p}
      \\
      p_{\infty} &=\frac{1}{1+\exp ((V+75) / 5.5)}
      \\
      \tau_{p} &=\frac{1}{\exp (-0.086 V-14.59)+\exp (0.0701 V-1.87)}
      \end{aligned}

  where :math:`\phi=1` is a temperature-dependent factor.

  Parameters
  ----------
  g_max : float
    The maximal conductance density (:math:`mS/cm^2`).
  E : float
    The reversal potential (mV).
  phi : float
    The temperature-dependent factor.

  References
  ----------
  .. [1] Huguenard, John R., and David A. McCormick. "Simulation of the currents
         involved in rhythmic oscillations in thalamic relay neurons." Journal
         of neurophysiology 68, no. 4 (1992): 1373-1383.

  """

  def __init__(
      self,
      size: Shape,
      g_max: Union[float, Tensor, Initializer, Callable]=10.,
      E: Union[float, Tensor, Initializer, Callable]=-90.,
      phi: Union[float, Tensor, Initializer, Callable]=1.,
      method: str = 'exp_auto',
      name: str = None
  ):
    super(Ih, self).__init__(size, name=name)

    # parameters
    self.phi = init_param(phi, self.num, allow_none=False)
    self.g_max = init_param(g_max, self.num, allow_none=False)
    self.E = init_param(E, self.num, allow_none=False)

    # variable
    self.p = bm.Variable(bm.zeros(self.num))

    # function
    self.integral = odeint(self.derivative, method=method)

  def derivative(self, p, t, V):
    p_inf = 1. / (1. + bm.exp((V + 75.) / 5.5))
    p_tau = 1. / (bm.exp(-0.086 * V - 14.59) + bm.exp(0.0701 * V - 1.87))
    return self.phi * (p_inf - p) / p_tau

  def reset(self, V):
    self.p.value = 1. / (1. + bm.exp((V + 75.) / 5.5))

  def update(self, t, dt, V):
    self.p.value = self.integral(self.p, t, V, dt=dt)

  def current(self, V):
    g = self.g_max * self.p
    return g * (self.E - V)



# -*- coding: utf-8 -*-

from typing import Union, Callable

import brainpy.math as bm
from brainpy.dyn.base import NeuGroup
from brainpy.initialize import OneInit, Uniform, Initializer, init_param
from brainpy.integrators.joint_eq import JointEq
from brainpy.integrators.ode import odeint
from brainpy.tools.checking import check_initializer
from brainpy.types import Shape, Tensor

__all__ = [
  'HH',
  'MorrisLecar',
  'PinskyRinzelModel',
  'WangBuzsakiModel',
]


class HH(NeuGroup):
  r"""Hodgkin–Huxley neuron model.

  **Model Descriptions**

  The Hodgkin-Huxley (HH; Hodgkin & Huxley, 1952) model [1]_ for the generation of
  the nerve action potential is one of the most successful mathematical models of
  a complex biological process that has ever been formulated. The basic concepts
  expressed in the model have proved a valid approach to the study of bio-electrical
  activity from the most primitive single-celled organisms such as *Paramecium*,
  right through to the neurons within our own brains.

  Mathematically, the model is given by,

  .. math::

      C \frac {dV} {dt} = -(\bar{g}_{Na} m^3 h (V &-E_{Na})
      + \bar{g}_K n^4 (V-E_K) + g_{leak} (V - E_{leak})) + I(t)

      \frac {dx} {dt} &= \alpha_x (1-x)  - \beta_x, \quad x\in {\rm{\{m, h, n\}}}

      &\alpha_m(V) = \frac {0.1(V+40)}{1-\exp(\frac{-(V + 40)} {10})}

      &\beta_m(V) = 4.0 \exp(\frac{-(V + 65)} {18})

      &\alpha_h(V) = 0.07 \exp(\frac{-(V+65)}{20})

      &\beta_h(V) = \frac 1 {1 + \exp(\frac{-(V + 35)} {10})}

      &\alpha_n(V) = \frac {0.01(V+55)}{1-\exp(-(V+55)/10)}

      &\beta_n(V) = 0.125 \exp(\frac{-(V + 65)} {80})

  The illustrated example of HH neuron model please see `this notebook <../neurons/HH_model.ipynb>`_.

  The Hodgkin–Huxley model can be thought of as a differential equation system with
  four state variables, :math:`V_{m}(t),n(t),m(t)`, and :math:`h(t)`, that change
  with respect to time :math:`t`. The system is difficult to study because it is a
  nonlinear system and cannot be solved analytically. However, there are many numeric
  methods available to analyze the system. Certain properties and general behaviors,
  such as limit cycles, can be proven to exist.

  *1. Center manifold*

  Because there are four state variables, visualizing the path in phase space can
  be difficult. Usually two variables are chosen, voltage :math:`V_{m}(t)` and the
  potassium gating variable :math:`n(t)`, allowing one to visualize the limit cycle.
  However, one must be careful because this is an ad-hoc method of visualizing the
  4-dimensional system. This does not prove the existence of the limit cycle.

  .. image:: ../../../../_static/Hodgkin_Huxley_Limit_Cycle.png
      :align: center

  A better projection can be constructed from a careful analysis of the Jacobian of
  the system, evaluated at the equilibrium point. Specifically, the eigenvalues of
  the Jacobian are indicative of the center manifold's existence. Likewise, the
  eigenvectors of the Jacobian reveal the center manifold's orientation. The
  Hodgkin–Huxley model has two negative eigenvalues and two complex eigenvalues
  with slightly positive real parts. The eigenvectors associated with the two
  negative eigenvalues will reduce to zero as time :math:`t` increases. The remaining
  two complex eigenvectors define the center manifold. In other words, the
  4-dimensional system collapses onto a 2-dimensional plane. Any solution
  starting off the center manifold will decay towards the *center manifold*.
  Furthermore, the limit cycle is contained on the center manifold.

  *2. Bifurcations*

  If the injected current :math:`I` were used as a bifurcation parameter, then the
  Hodgkin–Huxley model undergoes a Hopf bifurcation. As with most neuronal models,
  increasing the injected current will increase the firing rate of the neuron.
  One consequence of the Hopf bifurcation is that there is a minimum firing rate.
  This means that either the neuron is not firing at all (corresponding to zero
  frequency), or firing at the minimum firing rate. Because of the all-or-none
  principle, there is no smooth increase in action potential amplitude, but
  rather there is a sudden "jump" in amplitude. The resulting transition is
  known as a `canard <http://www.scholarpedia.org/article/Canards>`_.

  .. image:: ../../../../_static/Hodgkins_Huxley_bifurcation_by_I.gif
     :align: center

  The following image shows the bifurcation diagram of the Hodgkin–Huxley model
  as a function of the external drive :math:`I` [3]_. The green lines show the amplitude
  of a stable limit cycle and the blue lines indicate unstable limit-cycle behaviour,
  both born from Hopf bifurcations. The solid red line shows the stable fixed point
  and the black line shows the unstable fixed point.

  .. image:: ../../../../_static/Hodgkin_Huxley_bifurcation.png
     :align: center

  **Model Examples**

  .. plot::
    :include-source: True

    >>> import brainpy as bp
    >>> group = bp.dyn.HH(2)
    >>> runner = bp.dyn.DSRunner(group, monitors=['V'], inputs=('input', 10.))
    >>> runner.run(200.)
    >>> bp.visualize.line_plot(runner.mon.ts, runner.mon.V, show=True)

  .. plot::
    :include-source: True

    >>> import brainpy as bp
    >>> import matplotlib.pyplot as plt
    >>>
    >>> group = bp.dyn.HH(2)
    >>>
    >>> I1 = bp.inputs.spike_input(sp_times=[500., 550., 1000, 1030, 1060, 1100, 1200], sp_lens=5, sp_sizes=5., duration=2000, )
    >>> I2 = bp.inputs.spike_input(sp_times=[600.,       900, 950, 1500], sp_lens=5, sp_sizes=5., duration=2000, )
    >>> I1 += bp.math.random.normal(0, 3, size=I1.shape)
    >>> I2 += bp.math.random.normal(0, 3, size=I2.shape)
    >>> I = bp.math.stack((I1, I2), axis=-1)
    >>>
    >>> runner = bp.dyn.DSRunner(group, monitors=['V'], inputs=('input', I, 'iter'))
    >>> runner.run(2000.)
    >>>
    >>> fig, gs = bp.visualize.get_figure(1, 1, 3, 8)
    >>> fig.add_subplot(gs[0, 0])
    >>> plt.plot(runner.mon.ts, runner.mon.V[:, 0])
    >>> plt.plot(runner.mon.ts, runner.mon.V[:, 1] + 130)
    >>> plt.xlim(10, 2000)
    >>> plt.xticks([])
    >>> plt.yticks([])
    >>> plt.show()

  Parameters
  ----------
  size: sequence of int, int
    The size of the neuron group.
  ENa: float, JaxArray, ndarray, Initializer, callable
    The reversal potential of sodium. Default is 50 mV.
  gNa: float, JaxArray, ndarray, Initializer, callable
    The maximum conductance of sodium channel. Default is 120 msiemens.
  EK: float, JaxArray, ndarray, Initializer, callable
    The reversal potential of potassium. Default is -77 mV.
  gK: float, JaxArray, ndarray, Initializer, callable
    The maximum conductance of potassium channel. Default is 36 msiemens.
  EL: float, JaxArray, ndarray, Initializer, callable
    The reversal potential of learky channel. Default is -54.387 mV.
  gL: float, JaxArray, ndarray, Initializer, callable
    The conductance of learky channel. Default is 0.03 msiemens.
  V_th: float, JaxArray, ndarray, Initializer, callable
    The threshold of the membrane spike. Default is 20 mV.
  C: float, JaxArray, ndarray, Initializer, callable
    The membrane capacitance. Default is 1 ufarad.
  V_initializer: JaxArray, ndarray, Initializer, callable
    The initializer of membrane potential.
  m_initializer: JaxArray, ndarray, Initializer, callable
    The initializer of m channel.
  h_initializer: JaxArray, ndarray, Initializer, callable
    The initializer of h channel.
  n_initializer: JaxArray, ndarray, Initializer, callable
    The initializer of n channel.
  method: str
    The numerical integration method.
  name: str
    The group name.

  References
  ----------

  .. [1] Hodgkin, Alan L., and Andrew F. Huxley. "A quantitative description
         of membrane current and its application to conduction and excitation
         in nerve." The Journal of physiology 117.4 (1952): 500.
  .. [2] https://en.wikipedia.org/wiki/Hodgkin%E2%80%93Huxley_model
  .. [3] Ashwin, Peter, Stephen Coombes, and Rachel Nicks. "Mathematical
         frameworks for oscillatory network dynamics in neuroscience."
         The Journal of Mathematical Neuroscience 6, no. 1 (2016): 1-92.
  """

  def __init__(
      self,
      size: Shape,
      ENa: Union[float, Tensor, Initializer, Callable] = 50.,
      gNa: Union[float, Tensor, Initializer, Callable] = 120.,
      EK: Union[float, Tensor, Initializer, Callable] = -77.,
      gK: Union[float, Tensor, Initializer, Callable] = 36.,
      EL: Union[float, Tensor, Initializer, Callable] = -54.387,
      gL: Union[float, Tensor, Initializer, Callable] = 0.03,
      V_th: Union[float, Tensor, Initializer, Callable] = 20.,
      C: Union[float, Tensor, Initializer, Callable] = 1.0,
      V_initializer: Union[Initializer, Callable, Tensor] = Uniform(-70, -60.),
      m_initializer: Union[Initializer, Callable, Tensor] = OneInit(0.5),
      h_initializer: Union[Initializer, Callable, Tensor] = OneInit(0.6),
      n_initializer: Union[Initializer, Callable, Tensor] = OneInit(0.32),
      method: str = 'exp_auto',
      keep_size: bool = False,
      name: str = None
  ):
    # initialization
    super(HH, self).__init__(size=size, name=name)

    # parameters
    self.ENa = init_param(ENa, self.num, allow_none=False)
    self.EK = init_param(EK, self.num, allow_none=False)
    self.EL = init_param(EL, self.num, allow_none=False)
    self.gNa = init_param(gNa, self.num, allow_none=False)
    self.gK = init_param(gK, self.num, allow_none=False)
    self.gL = init_param(gL, self.num, allow_none=False)
    self.C = init_param(C, self.num, allow_none=False)
    self.V_th = init_param(V_th, self.num, allow_none=False)

    # initializers
    check_initializer(m_initializer, 'm_initializer', allow_none=False)
    check_initializer(h_initializer, 'h_initializer', allow_none=False)
    check_initializer(n_initializer, 'n_initializer', allow_none=False)
    check_initializer(V_initializer, 'V_initializer', allow_none=False)
    self._m_initializer = m_initializer
    self._h_initializer = h_initializer
    self._n_initializer = n_initializer
    self._V_initializer = V_initializer

    # variables
    self.m = bm.Variable(init_param(self._m_initializer, (self.num,)))
    self.h = bm.Variable(init_param(self._h_initializer, (self.num,)))
    self.n = bm.Variable(init_param(self._n_initializer, (self.num,)))
    self.V = bm.Variable(init_param(self._V_initializer, (self.num,)))
    self.input = bm.Variable(bm.zeros(self.num))
    self.spike = bm.Variable(bm.zeros(self.num, dtype=bool))

    # integral
    self.integral = odeint(method=method, f=self.derivative)

  def reset(self):
    self.m.value = init_param(self._m_initializer, (self.num,))
    self.h.value = init_param(self._h_initializer, (self.num,))
    self.n.value = init_param(self._n_initializer, (self.num,))
    self.V.value = init_param(self._V_initializer, (self.num,))
    self.input[:] = 0
    self.spike[:] = False

  def dm(self, m, t, V):
    alpha = 0.1 * (V + 40) / (1 - bm.exp(-(V + 40) / 10))
    beta = 4.0 * bm.exp(-(V + 65) / 18)
    dmdt = alpha * (1 - m) - beta * m
    return dmdt

  def dh(self, h, t, V):
    alpha = 0.07 * bm.exp(-(V + 65) / 20.)
    beta = 1 / (1 + bm.exp(-(V + 35) / 10))
    dhdt = alpha * (1 - h) - beta * h
    return dhdt

  def dn(self, n, t, V):
    alpha = 0.01 * (V + 55) / (1 - bm.exp(-(V + 55) / 10))
    beta = 0.125 * bm.exp(-(V + 65) / 80)
    dndt = alpha * (1 - n) - beta * n
    return dndt

  def dV(self, V, t, m, h, n, I_ext):
    I_Na = (self.gNa * m ** 3.0 * h) * (V - self.ENa)
    I_K = (self.gK * n ** 4.0) * (V - self.EK)
    I_leak = self.gL * (V - self.EL)
    dVdt = (- I_Na - I_K - I_leak + I_ext) / self.C
    return dVdt

  @property
  def derivative(self):
    return JointEq([self.dV, self.dm, self.dh, self.dn])

  def update(self, t, dt):
    V, m, h, n = self.integral(self.V, self.m, self.h, self.n, t, self.input, dt=dt)
    self.spike.value = bm.logical_and(self.V < self.V_th, V >= self.V_th)
    self.V.value = V
    self.m.value = m
    self.h.value = h
    self.n.value = n
    self.input[:] = 0.



class MorrisLecar(NeuGroup):
  r"""The Morris-Lecar neuron model.

  **Model Descriptions**

  The Morris-Lecar model [4]_ (Also known as :math:`I_{Ca}+I_K`-model)
  is a two-dimensional "reduced" excitation model applicable to
  systems having two non-inactivating voltage-sensitive conductances.
  This model was named after Cathy Morris and Harold Lecar, who
  derived it in 1981. Because it is two-dimensional, the Morris-Lecar
  model is one of the favorite conductance-based models in computational neuroscience.

  The original form of the model employed an instantaneously
  responding voltage-sensitive Ca2+ conductance for excitation and a delayed
  voltage-dependent K+ conductance for recovery. The equations of the model are:

  .. math::

      \begin{aligned}
      C\frac{dV}{dt} =& -  g_{Ca} M_{\infty} (V - V_{Ca})- g_{K} W(V - V_{K}) -
                        g_{Leak} (V - V_{Leak}) + I_{ext} \\
      \frac{dW}{dt} =& \frac{W_{\infty}(V) - W}{ \tau_W(V)}
      \end{aligned}

  Here, :math:`V` is the membrane potential, :math:`W` is the "recovery variable",
  which is almost invariably the normalized :math:`K^+`-ion conductance, and
  :math:`I_{ext}` is the applied current stimulus.

  **Model Examples**

  .. plot::
    :include-source: True

    >>> import brainpy as bp
    >>>
    >>> group = bp.dyn.MorrisLecar(1)
    >>> runner = bp.dyn.DSRunner(group, monitors=['V', 'W'], inputs=('input', 100.))
    >>> runner.run(1000)
    >>>
    >>> fig, gs = bp.visualize.get_figure(2, 1, 3, 8)
    >>> fig.add_subplot(gs[0, 0])
    >>> bp.visualize.line_plot(runner.mon.ts, runner.mon.W, ylabel='W')
    >>> fig.add_subplot(gs[1, 0])
    >>> bp.visualize.line_plot(runner.mon.ts, runner.mon.V, ylabel='V', show=True)


  **Model Parameters**

  ============= ============== ======== =======================================================
  **Parameter** **Init Value** **Unit** **Explanation**
  ------------- -------------- -------- -------------------------------------------------------
  V_Ca          130            mV       Equilibrium potentials of Ca+.(mV)
  g_Ca          4.4            \        Maximum conductance of corresponding Ca+.(mS/cm2)
  V_K           -84            mV       Equilibrium potentials of K+.(mV)
  g_K           8              \        Maximum conductance of corresponding K+.(mS/cm2)
  V_Leak        -60            mV       Equilibrium potentials of leak current.(mV)
  g_Leak        2              \        Maximum conductance of leak current.(mS/cm2)
  C             20             \        Membrane capacitance.(uF/cm2)
  V1            -1.2           \        Potential at which M_inf = 0.5.(mV)
  V2            18             \        Reciprocal of slope of voltage dependence of M_inf.(mV)
  V3            2              \        Potential at which W_inf = 0.5.(mV)
  V4            30             \        Reciprocal of slope of voltage dependence of W_inf.(mV)
  phi           0.04           \        A temperature factor. (1/s)
  V_th          10             mV       The spike threshold.
  ============= ============== ======== =======================================================

  References
  ----------

  .. [4] Meier, Stephen R., Jarrett L. Lancaster, and Joseph M. Starobin.
         "Bursting regimes in a reaction-diffusion system with action
         potential-dependent equilibrium." PloS one 10.3 (2015):
         e0122401.
  .. [5] http://www.scholarpedia.org/article/Morris-Lecar_model
  .. [6] https://en.wikipedia.org/wiki/Morris%E2%80%93Lecar_model
  """

  def __init__(
      self,
      size: Shape,
      V_Ca: Union[float, Tensor, Initializer, Callable] = 130.,
      g_Ca: Union[float, Tensor, Initializer, Callable] = 4.4,
      V_K: Union[float, Tensor, Initializer, Callable] = -84.,
      g_K: Union[float, Tensor, Initializer, Callable] = 8.,
      V_leak: Union[float, Tensor, Initializer, Callable] = -60.,
      g_leak: Union[float, Tensor, Initializer, Callable] = 2.,
      C: Union[float, Tensor, Initializer, Callable] = 20.,
      V1: Union[float, Tensor, Initializer, Callable] = -1.2,
      V2: Union[float, Tensor, Initializer, Callable] = 18.,
      V3: Union[float, Tensor, Initializer, Callable] = 2.,
      V4: Union[float, Tensor, Initializer, Callable] = 30.,
      phi: Union[float, Tensor, Initializer, Callable] = 0.04,
      V_th: Union[float, Tensor, Initializer, Callable] = 10.,
      W_initializer: Union[Callable, Initializer, Tensor] = OneInit(0.02),
      V_initializer: Union[Callable, Initializer, Tensor] = Uniform(-70., -60.),
      method: str = 'exp_auto',
      keep_size: bool = False,
      name: str = None
  ):
    # initialization
    super(MorrisLecar, self).__init__(size=size, name=name)

    # params
    self.V_Ca = init_param(V_Ca, self.num, allow_none=False)
    self.g_Ca = init_param(g_Ca, self.num, allow_none=False)
    self.V_K = init_param(V_K, self.num, allow_none=False)
    self.g_K = init_param(g_K, self.num, allow_none=False)
    self.V_leak = init_param(V_leak, self.num, allow_none=False)
    self.g_leak = init_param(g_leak, self.num, allow_none=False)
    self.C = init_param(C, self.num, allow_none=False)
    self.V1 = init_param(V1, self.num, allow_none=False)
    self.V2 = init_param(V2, self.num, allow_none=False)
    self.V3 = init_param(V3, self.num, allow_none=False)
    self.V4 = init_param(V4, self.num, allow_none=False)
    self.phi = init_param(phi, self.num, allow_none=False)
    self.V_th = init_param(V_th, self.num, allow_none=False)

    # initializers
    check_initializer(V_initializer, 'V_initializer', allow_none=False)
    check_initializer(W_initializer, 'W_initializer', allow_none=False)
    self._W_initializer = W_initializer
    self._V_initializer = V_initializer

    # variables
    self.W = bm.Variable(init_param(W_initializer, (self.num,)))
    self.V = bm.Variable(init_param(V_initializer, (self.num,)))
    self.input = bm.Variable(bm.zeros(self.num))
    self.spike = bm.Variable(bm.zeros(self.num, dtype=bool))

    # integral
    self.integral = odeint(method=method, f=self.derivative)

  def reset(self):
    self.W.value = init_param(self._W_initializer, (self.num,))
    self.V.value = init_param(self._V_initializer, (self.num,))
    self.input.value = bm.zeros(self.num)
    self.spike.value = bm.zeros(self.num, dtype=bool)

  def dV(self, V, t, W, I_ext):
    M_inf = (1 / 2) * (1 + bm.tanh((V - self.V1) / self.V2))
    I_Ca = self.g_Ca * M_inf * (V - self.V_Ca)
    I_K = self.g_K * W * (V - self.V_K)
    I_Leak = self.g_leak * (V - self.V_leak)
    dVdt = (- I_Ca - I_K - I_Leak + I_ext) / self.C
    return dVdt

  def dW(self, W, t, V):
    tau_W = 1 / (self.phi * bm.cosh((V - self.V3) / (2 * self.V4)))
    W_inf = (1 / 2) * (1 + bm.tanh((V - self.V3) / self.V4))
    dWdt = (W_inf - W) / tau_W
    return dWdt

  @property
  def derivative(self):
    return JointEq([self.dV, self.dW])

  def update(self, t, dt):
    V, self.W.value = self.integral(self.V, self.W, t, self.input, dt=dt)
    spike = bm.logical_and(self.V < self.V_th, V >= self.V_th)
    self.V.value = V
    self.spike.value = spike
    self.input[:] = 0.


class PinskyRinzelModel(NeuGroup):
  r"""The Pinsky and Rinsel (1994) model.

  The Pinsky and Rinsel (1994) model [7]_ is a 2-compartment (soma and dendrite),
  conductance-based (Hodgin-Huxley type) model of a hippocampal CA3 pyramidal
  neuron. It is a reduced version of an earlier, 19-compartment model by
  Traub, et. al. (1991) [8]_. This model demonstrates how similar qualitative
  and quantitative spiking behaviors can be obtained despite the reduction
  in model complexity.

  Specifically, this model demonstrates calcium bursting behavior and how
  the 'ping-pong' interplay between somatic and dendritic currents results
  in a complex shape of the burst.

  .. image:: ../../../../_static/Pinsky-Rinzel-model-illustration.png
      :align: center

  Mathematically, the model is given by:

  .. math::

     \begin{aligned}
    &\mathrm{C}_{\mathrm{m}} \mathrm{V}_{\mathrm{s}}^{\prime}=-\mathrm{I}_{\mathrm{Leak}}-\mathrm{I}_{\mathrm{Na}}-\mathrm{I}_{\mathrm{K}_{\mathrm{DR}}}-\frac{\mathrm{I}_{\mathrm{DS}}}{\mathrm{p}}+\frac{\mathrm{I}_{\mathrm{S}_{\mathrm{app}}}}{\mathrm{p}} \\
    &\mathrm{C}_{\mathrm{m}} \mathrm{V}_{\mathrm{d}}^{\prime}=-\mathrm{I}_{\mathrm{Leak}}-\mathrm{I}_{\mathrm{Ca}}-\mathrm{I}_{\mathrm{K}_{\mathrm{Ca}}}-\mathrm{I}_{\mathrm{K}_{\mathrm{AHP}}}+\frac{\mathrm{I}_{\mathrm{SD}}}{(1-\mathrm{p})}+\frac{\mathrm{I}_{\mathrm{D}_{\mathrm{app}}}}{(1-\mathrm{p})} \\
    &\frac{\mathrm{dCa}}{\mathrm{dt}}=-0.13 \mathrm{I}_{\mathrm{Ca}}-0.075 \mathrm{Ca}
    \end{aligned}

  The currents of the model are functions of potentials as follows:

  .. math::

     \begin{aligned}
      \mathrm{I}_{\mathrm{Na}} &=\mathrm{g}_{\mathrm{Na}} m_{\infty}^{2}\left(\mathrm{~V}_{\mathrm{s}}\right) h\left(\mathrm{~V}_{\mathrm{s}}-\mathrm{V}_{\mathrm{Na}}\right) \\
      \mathrm{I}_{\mathrm{K}_{\mathrm{DR}}} &=\mathrm{g}_{\mathrm{K}_{\mathrm{DR}}} n\left(\mathrm{~V}_{\mathrm{s}}-\mathrm{V}_{\mathrm{K}}\right) \\
      \mathrm{I}_{\mathrm{Ca}} &=\mathrm{g}_{\mathrm{Ca}}{ }^{2}\left(\mathrm{~V}_{\mathrm{d}}-\mathrm{V}_{\mathrm{N}}\right) \\
      \mathrm{I}_{\mathrm{K}_{\mathrm{Ca}}} &=\mathrm{g}_{\mathrm{k}_{\mathrm{Ca}}} C \chi(\mathrm{Ca})\left(\mathrm{V}_{\mathrm{d}}-\mathrm{V}_{\mathrm{Ca}}\right) \\
      \mathrm{I}_{\mathrm{K}_{\mathrm{AHP}}} &=\mathrm{g}_{\mathrm{K}_{\mathrm{AHP}}} q\left(\mathrm{~V}_{\mathrm{d}}-\mathrm{V}_{\mathrm{K}}\right) \\
      \mathrm{I}_{\mathrm{SD}} &=-\mathrm{I}_{\mathrm{DS}}=\mathrm{g}_{\mathrm{c}}\left(\mathrm{V}_{\mathrm{d}}-\mathrm{V}_{\mathrm{s}}\right) \\
      \mathrm{I}_{\mathrm{Leak}} &=\mathrm{g}_{\mathrm{L}}\left(\mathrm{V}-\mathrm{V}_{\mathrm{L}}\right)
      \end{aligned}

  The activation and inactivation variables should satisfy these equations

  .. math::

     \begin{aligned}
    \omega^{\prime}(\mathrm{V}) &=\frac{\omega_{\infty}(\mathrm{V})-\omega}{\tau_{\omega}(\mathrm{V})} \\
    \omega_{\infty}(\mathrm{V}) &=\frac{\alpha_{\omega}(\mathrm{V})}{\alpha_{\omega}(\mathrm{V})+\beta_{\omega}(\mathrm{V})} \\
    \tau_{\omega}(\mathrm{V}) &=\frac{1}{\alpha_{\omega}(\mathrm{V})+\beta_{\omega}(\mathrm{V})}
    \end{aligned}

  where, independently, we consider :math:`\omega = h, n, s, m, c, q`.

  The rate functions are defined as follows

  .. math::

     \begin{aligned}
    \alpha_{m}\left(\mathrm{~V}_{\mathrm{s}}\right) &=\frac{0.32\left(-46.9-\mathrm{V}_{\mathrm{s}}\right)}{\exp \left(\frac{-46.9-\mathrm{V}_{\mathrm{s}}}{4}\right)-1} \\
    \beta_{m}\left(\mathrm{~V}_{\mathrm{s}}\right) &=\frac{0.28\left(\mathrm{~V}_{\mathrm{s}}+19.9\right)}{\exp \left(\frac{\mathrm{V}_{\mathrm{s}}+19.9}{5}\right)-1}, \\
    \alpha_{n}\left(\mathrm{~V}_{\mathrm{s}}\right) &=\frac{0.016\left(-24.9-\mathrm{V}_{\mathrm{s}}\right)}{\exp \left(\frac{-24.9-\mathrm{V}_{\mathrm{s}}}{5}\right)-1} \\
    \beta_{n}\left(\mathrm{~V}_{\mathrm{s}}\right) &=0.25 \exp \left(-1-0.025 \mathrm{~V}_{\mathrm{s}}\right) \\
    \alpha_{h}\left(\mathrm{~V}_{\mathrm{s}}\right) &=0.128 \exp \left(\frac{-43-\mathrm{V}_{\mathrm{s}}}{18}\right) \\
    \beta_{h}\left(\mathrm{~V}_{\mathrm{s}}\right) &=\frac{4}{1+\exp \left(\frac{\left(-20-\mathrm{V}_{\mathrm{s}}\right.}{5}\right)}, \\
    \alpha_{s}\left(\mathrm{~V}_{\mathrm{d}}\right) &=\frac{1.6}{1+\exp \left(-0.072\left(\mathrm{~V}_{\mathrm{d}}-5\right)\right)} \\
    \beta_{s}\left(\mathrm{~V}_{\mathrm{d}}\right) &=\frac{0.02\left(\mathrm{~V}_{\mathrm{d}}+8.9\right)}{\exp \left(\frac{\left(\mathrm{V}_{\mathrm{d}}+8.9\right)}{5}\right)-1}, \\
    \alpha_{C}\left(\mathrm{~V}_{\mathrm{d}}\right) &=\frac{\left(1-H\left(\mathrm{~V}_{\mathrm{d}}+10\right)\right) \exp \left(\frac{\left(\mathrm{V}_{\mathrm{d}}+50\right)}{11}-\frac{\left(\mathrm{V}_{\mathrm{d}}+53.5\right)}{27}\right)}{18.975}+H\left(\mathrm{~V}_{\mathrm{d}}+10\right)\left(2 \exp \left(\frac{\left(-53.5-\mathrm{V}_{\mathrm{d}}\right.}{27}\right)\right) \\
    \beta_{C}\left(\mathrm{~V}_{\mathrm{d}}\right) &=\left(1-H\left(\mathrm{~V}_{\mathrm{d}}+10\right)\right)\left(2 \exp \left(\frac{\left(-53.5-\mathrm{V}_{\mathrm{d}}\right)}{27}\right)-\alpha_{c}\left(\mathrm{~V}_{\mathrm{d}}\right)\right) \\
    \alpha_{q}(\mathrm{Ca}) &=\min (0.00002 \mathrm{Ca}, 0.01) \\
    \beta_{q}(\mathrm{Ca}) &=0.001 \\
    \chi(\mathrm{Ca}) &=\min \left(\frac{\mathrm{Ca}}{250}, 1\right)
    \end{aligned}

  The standard values of the parameters are given below. The maximal conductances
  (in :math:`\mathrm{mS} / \mathrm{cm}^{2}`) are
  :math:`\bar{g}_{L}=0.1`, :math:`\bar{g}_{\mathrm{Na}}=30`,
  :math:`\bar{g}_{\mathrm{K}-\mathrm{DR}}=15`,
  :math:`\bar{g}_{\mathrm{Ca}}=10`,
  :math:`\bar{g}_{\mathrm{K}-\mathrm{AHP}}=0.8`,
  :math:`\bar{g}_{\mathrm{K}-\mathrm{C}}=15`,
  :math:`\bar{g}_{\mathrm{NMDA}}=0.0` and
  :math:`\bar{g}_{\mathrm{AMPA}}=0.0`.
  The reversal potentials (in :math:`\mathrm{mV}` ) are
  :math:`V_{\mathrm{Na}}=120, V_{\mathrm{C}}=140, V_{\mathrm{K}}=-15 \mathrm{mV})`
  are :math:`V_{\mathrm{Na}}=120, V_{\mathrm{Ca}}=140, V_{\mathrm{K}}=-15, $V_{L}=0`
  and :math:`V_{\text {Syn }}=60`. The applied currents
  (in :math:`\mu \mathrm{A} / \mathrm{cm}^{2}` ) are :math:`I_{s}=-0.5` and :math:`I_{d}=0.0`.
  The coupling parameters are :math:`g_{c}=2.1 \mathrm{mS} / \mathrm{cm}^{2}` and
  :math:`p=0.5`. The capacitance, :math:`C_{M}`, is
  :math:`3 \mu \mathrm{F} / \mathrm{cm}^{2}` and :math:`\chi(C a)=\min (C a / 250,1)`.
  Values for these parameters, and these function definitions, are taken from Traub et al, 1991.


  Parameters
  ----------
  size: sequence of int, int
    The size of the neuron group.
  gNa: float, JaxArray, ndarray, Initializer, callable
    The maximum conductance of sodium channel.
  gK: float, JaxArray, ndarray, Initializer, callable
    The maximum conductance of potassium delayed-rectifier channel.
  gCa: float, JaxArray, ndarray, Initializer, callable
    The maximum conductance of calcium channel.
  gAHP: float, JaxArray, ndarray, Initializer, callable
    The maximum conductance of potassium after-hyper-polarization channel.
  gC: float, JaxArray, ndarray, Initializer, callable
    The maximum conductance of calcium activated potassium channel.
  gL: float, JaxArray, ndarray, Initializer, callable
    The conductance of leaky channel.
  ENa: float, JaxArray, ndarray, Initializer, callable
    The reversal potential of sodium channel.
  EK: float, JaxArray, ndarray, Initializer, callable
    The reversal potential of potassium delayed-rectifier channel.
  ECa: float, JaxArray, ndarray, Initializer, callable
    The reversal potential of calcium channel.
  EL: float, JaxArray, ndarray, Initializer, callable
    The reversal potential of leaky channel.
  gc: float, JaxArray, ndarray, Initializer, callable
    The coupling strength between the soma and dendrite.
  V_th: float, JaxArray, ndarray, Initializer, callable
    The threshold of the membrane spike.
  Cm: float, JaxArray, ndarray, Initializer, callable
    The threshold of the membrane spike.
  A: float, JaxArray, ndarray, Initializer, callable
    The total cell membrane area, which is normalized to 1.
  p: float, JaxArray, ndarray, Initializer, callable
    The proportion of cell area taken up by the soma.
  Vs_initializer: JaxArray, ndarray, Initializer, callable
    The initializer of somatic membrane potential.
  Vd_initializer: JaxArray, ndarray, Initializer, callable
    The initializer of dendritic membrane potential.
  Ca_initializer: JaxArray, ndarray, Initializer, callable
    The initializer of Calcium concentration.
  method: str
    The numerical integration method.
  name: str
    The group name.

  References
  ----------
  .. [7] Pinsky, Paul F., and John Rinzel. "Intrinsic and network
         rhythmogenesis in a reduced Traub model for CA3 neurons."
         Journal of computational neuroscience 1.1 (1994): 39-60.
  .. [8] Traub, R. D., Wong, R. K., Miles, R., & Michelson, H. (1991).
         A model of a CA3 hippocampal pyramidal neuron incorporating
         voltage-clamp data on intrinsic conductances. Journal of
         neurophysiology, 66(2), 635-650.
  """

  def __init__(
      self,
      size: Shape,
      # maximum conductance
      gNa: Union[float, Tensor, Initializer, Callable] = 30.,
      gK: Union[float, Tensor, Initializer, Callable] = 15.,
      gCa: Union[float, Tensor, Initializer, Callable] = 10.,
      gAHP: Union[float, Tensor, Initializer, Callable] = 0.8,
      gC: Union[float, Tensor, Initializer, Callable] = 15.,
      gL: Union[float, Tensor, Initializer, Callable] = 0.1,
      # reversal potential
      ENa: Union[float, Tensor, Initializer, Callable] = 60.,
      EK: Union[float, Tensor, Initializer, Callable] = -75.,
      ECa: Union[float, Tensor, Initializer, Callable] = 80.,
      EL: Union[float, Tensor, Initializer, Callable] = -60.,
      # other parameters
      gc: Union[float, Tensor, Initializer, Callable] = 2.1,
      V_th: Union[float, Tensor, Initializer, Callable] = 20.,
      Cm: Union[float, Tensor, Initializer, Callable] = 3.0,
      p: Union[float, Tensor, Initializer, Callable] = 0.5,
      A: Union[float, Tensor, Initializer, Callable] = 1.,
      # initializers
      Vs_initializer: Union[Initializer, Callable, Tensor] = OneInit(-64.6),
      Vd_initializer: Union[Initializer, Callable, Tensor] = OneInit(-64.5),
      Ca_initializer: Union[Initializer, Callable, Tensor] = OneInit(0.2),
      # others
      method: str = 'exp_auto',
      keep_size: bool = False,
      name: str = None,
  ):
    # initialization
    super(PinskyRinzelModel, self).__init__(size=size, name=name)

    # conductance parameters
    self.gAHP = init_param(gAHP, self.num, allow_none=False)
    self.gCa = init_param(gCa, self.num, allow_none=False)
    self.gNa = init_param(gNa, self.num, allow_none=False)
    self.gK = init_param(gK, self.num, allow_none=False)
    self.gL = init_param(gL, self.num, allow_none=False)
    self.gC = init_param(gC, self.num, allow_none=False)

    # reversal potential parameters
    self.ENa = init_param(ENa, self.num, allow_none=False)
    self.ECa = init_param(ECa, self.num, allow_none=False)
    self.EK = init_param(EK, self.num, allow_none=False)
    self.EL = init_param(EL, self.num, allow_none=False)

    # other neuronal parameters
    self.V_th = init_param(V_th, self.num, allow_none=False)
    self.Cm = init_param(Cm, self.num, allow_none=False)
    self.gc = init_param(gc, self.num, allow_none=False)
    self.p = init_param(p, self.num, allow_none=False)
    self.A = init_param(A, self.num, allow_none=False)

    # initializers
    check_initializer(Vs_initializer, 'Vs_initializer', allow_none=False)
    check_initializer(Vd_initializer, 'Vd_initializer', allow_none=False)
    check_initializer(Ca_initializer, 'Ca_initializer', allow_none=False)
    self._Vs_initializer = Vs_initializer
    self._Vd_initializer = Vd_initializer
    self._Ca_initializer = Ca_initializer

    # variables
    self.Vs = bm.Variable(init_param(self._Vs_initializer, (self.num,)))
    self.Vd = bm.Variable(init_param(self._Vd_initializer, (self.num,)))
    self.Ca = bm.Variable(init_param(self._Ca_initializer, (self.num,)))
    self.h = bm.Variable(self.inf_h(self.Vs))
    self.n = bm.Variable(self.inf_n(self.Vs))
    self.s = bm.Variable(self.inf_s(self.Vd))
    self.c = bm.Variable(self.inf_c(self.Vd))
    self.q = bm.Variable(self.inf_q(self.Ca))
    self.Id = bm.Variable(bm.zeros((self.num,)))  # input to soma
    self.Is = bm.Variable(bm.zeros((self.num,)))  # input to dendrite
    # self.spike = bm.Variable(bm.zeros(self.num, dtype=bool))

    # integral
    self.integral = odeint(method=method, f=self.derivative)

  def reset(self):
    self.Vd.value = init_param(self._Vd_initializer, (self.num,))
    self.Vs.value = init_param(self._Vs_initializer, (self.num,))
    self.Ca.value = init_param(self._Ca_initializer, (self.num,))
    self.h.value = self.inf_h(self.Vs)
    self.n.value = self.inf_n(self.Vs)
    self.s.value = self.inf_s(self.Vd)
    self.c.value = self.inf_c(self.Vd)
    self.q.value = self.inf_q(self.Ca)
    self.Id[:] = 0
    self.Is[:] = 0
    # self.spike[:] = False

  def dCa(self, Ca, t, s, Vd):
    ICa = self.gCa * s * s * (Vd - self.ECa)
    return -0.13 * ICa - 0.075 * Ca

  def dh(self, h, t, Vs): return self.alpha_h(Vs) * (1 - h) - self.beta_h(Vs) * h

  def dn(self, n, t, Vs): return self.alpha_n(Vs) * (1 - n) - self.beta_n(Vs) * n

  def ds(self, s, t, Vd): return self.alpha_s(Vd) * (1 - s) - self.beta_s(Vd) * s

  def dc(self, c, t, Vd): return self.alpha_c(Vd) * (1 - c) - self.beta_c(Vd) * c

  def dq(self, q, t, Ca): return self.alpha_q(Ca) * (1 - q) - self.beta_q(Ca) * q

  def dVs(self, Vs, t, h, n, Vd):
    I_Na = (self.gNa * self.inf_m(Vs) ** 2 * h) * (Vs - self.ENa)
    I_KDR = (self.gK * n) * (Vs - self.EK)
    I_leak = self.gL * (Vs - self.EL)
    I_gj = self.gc / self.p * (Vd - Vs)
    dVdt = (- I_Na - I_KDR - I_leak + I_gj + self.Is / self.p) / self.Cm
    return dVdt

  def dVd(self, Vd, t, s, q, c, Ca, Vs):
    I_leak = self.gL * (Vd - self.EL)
    I_Ca = self.gCa * s * s * (Vd - self.ECa)
    I_AHP = self.gAHP * q * (Vd - self.EK)
    I_C = self.gC * bm.minimum(Ca / 250., 1.) * (Vd - self.EK)
    p = 1 - self.p
    I_gj = self.gc / p * (Vs - Vd)
    dVdt = (- I_leak - I_Ca - I_AHP - I_C + I_gj + self.Id / p) / self.Cm
    return dVdt

  @property
  def derivative(self):
    return JointEq([self.dVs, self.dVd, self.dCa, self.dh, self.dn, self.ds, self.dc, self.dq])

  def update(self, t, dt):
    Vs, Vd, Ca, h, n, s, c, q = self.integral(Vs=self.Vs.value,
                                              Vd=self.Vd.value,
                                              Ca=self.Ca.value,
                                              h=self.h.value,
                                              n=self.n.value,
                                              s=self.s.value,
                                              c=self.c.value,
                                              q=self.q.value,
                                              t=t,
                                              dt=dt)
    self.Vs.value = Vs
    self.Vd.value = Vd
    self.Ca.value = Ca
    self.h.value = h
    self.n.value = n
    self.s.value = s
    self.c.value = c
    self.q.value = q
    self.Id[:] = 0.
    self.Is[:] = 0.

  def alpha_m(self, Vs): return 0.32 * (13.1 - (Vs + 60.)) / (bm.exp((13.1 - (Vs + 60.)) / 4.) - 1.)

  def beta_m(self, Vs): return 0.28 * ((Vs + 60.) - 40.1) / (bm.exp(((Vs + 60.) - 40.1) / 5.) - 1.)

  def inf_m(self, Vs):
    alpha = self.alpha_m(Vs)
    beta = self.beta_m(Vs)
    return alpha / (alpha + beta)

  def alpha_n(self, Vs): return 0.016 * (35.1 - (Vs + 60.)) / (bm.exp((35.1 - (Vs + 60.)) / 5) - 1)

  def beta_n(self, Vs): return 0.25 * bm.exp(0.5 - 0.025 * (Vs + 60.))

  def inf_n(self, Vs):
    alpha = self.alpha_n(Vs)
    beta = self.beta_n(Vs)
    return alpha / (alpha + beta)

  def alpha_h(self, Vs): return 0.128 * bm.exp((17. - (Vs + 60.)) / 18.)

  def beta_h(self, Vs): return 4. / (1 + bm.exp((40. - (Vs + 60.)) / 5))

  def inf_h(self, Vs):
    alpha = self.alpha_h(Vs)
    beta = self.beta_h(Vs)
    return alpha / (alpha + beta)

  def alpha_s(self, Vd): return 1.6 / (1 + bm.exp(-0.072 * ((Vd + 60.) - 65.)))

  def beta_s(self, Vd): return 0.02 * ((Vd + 60.) - 51.1) / (bm.exp(((Vd + 60.) - 51.1) / 5.) - 1.)

  def inf_s(self, Vd):
    alpha = self.alpha_s(Vd)
    beta = self.beta_s(Vd)
    return alpha / (alpha + beta)

  def alpha_c(self, Vd):
    return bm.where((Vd + 60.) <= 50.,
                    (bm.exp(((Vd + 60.) - 10.) / 11.) - bm.exp(((Vd + 60.) - 6.5) / 27.)) / 18.975,
                    2. * bm.exp((6.5 - (Vd + 60.)) / 27.))

  def beta_c(self, Vd):
    alpha_c = (bm.exp(((Vd + 60.) - 10.) / 11.) - bm.exp(((Vd + 60.) - 6.5) / 27.)) / 18.975
    return bm.where((Vd + 60.) <= 50., 2. * bm.exp((6.5 - (Vd + 60.)) / 27.) - alpha_c, 0.)

  def inf_c(self, Vd):
    alpha_c = self.alpha_c(Vd)
    beta_c = self.beta_c(Vd)
    return alpha_c / (alpha_c + beta_c)

  def alpha_q(self, Ca): return bm.minimum(2e-5 * Ca, 1e-2)

  def beta_q(self, Ca): return 1e-3

  def inf_q(self, Ca):
    alpha = self.alpha_q(Ca)
    beta = self.beta_q(Ca)
    return alpha / (alpha + beta)


class WangBuzsakiModel(NeuGroup):
  r"""Wang-Buzsaki model [9]_, an implementation of a modified Hodgkin-Huxley model.

  Each model is described by a single compartment and obeys the current balance equation:

  .. math::

      C_{m} \frac{d V}{d t}=-I_{\mathrm{Na}}-I_{\mathrm{K}}-I_{\mathrm{L}}-I_{\mathrm{syn}}+I_{\mathrm{app}}

  where :math:`C_{m}=1 \mu \mathrm{F} / \mathrm{cm}^{2}` and :math:`I_{\mathrm{app}}` is the
  injected current (in :math:`\mu \mathrm{A} / \mathrm{cm}^{2}` ). The leak current
  :math:`I_{\mathrm{L}}=g_{\mathrm{L}}\left(V-E_{\mathrm{L}}\right)` has a conductance
  :math:`g_{\mathrm{L}}=0.1 \mathrm{mS} / \mathrm{cm}^{2}`, so that the passive time constant
  :math:`\tau_{0}=C_{m} / g_{\mathrm{L}}=10 \mathrm{msec} ; E_{\mathrm{L}}=-65 \mathrm{mV}`.

  The spike-generating :math:`\mathrm{Na}^{+}` and :math:`\mathrm{K}^{+}` voltage-dependent ion
  currents :math:`\left(I_{\mathrm{Na}}\right.` and :math:`I_{\mathrm{K}}` ) are of the
  Hodgkin-Huxley type (Hodgkin and Huxley, 1952). The transient sodium current
  :math:`I_{\mathrm{Na}}=g_{\mathrm{Na}} m_{\infty}^{3} h\left(V-E_{\mathrm{Na}}\right)`,
  where the activation variable :math:`m` is assumed fast and substituted by its steady-state
  function :math:`m_{\infty}=\alpha_{m} /\left(\alpha_{m}+\beta_{m}\right)` ;
  :math:`\alpha_{m}(V)=-0.1(V+35) /(\exp (-0.1(V+35))-1), \beta_{m}(V)=4 \exp (-(V+60) / 18)`.
  The inactivation variable :math:`h` obeys a first-order kinetics:

  .. math::

      \frac{d h}{d t}=\phi\left(\alpha_{h}(1-h)-\beta_{h} h\right)

  where :math:`\alpha_{h}(V)=0.07 \exp (-(V+58) / 20)` and
  :math:`\beta_{h}(V)=1 /(\exp (-0.1(V+28)) +1) \cdot g_{\mathrm{Na}}=35 \mathrm{mS} / \mathrm{cm}^{2}` ;
  :math:`E_{\mathrm{Na}}=55 \mathrm{mV}, \phi=5 .`

  The delayed rectifier :math:`I_{\mathrm{K}}=g_{\mathrm{K}} n^{4}\left(V-E_{\mathrm{K}}\right)`,
  where the activation variable :math:`n` obeys the following equation:

  .. math::

     \frac{d n}{d t}=\phi\left(\alpha_{n}(1-n)-\beta_{n} n\right)

  with :math:`\alpha_{n}(V)=-0.01(V+34) /(\exp (-0.1(V+34))-1)` and
  :math:`\beta_{n}(V)=0.125\exp (-(V+44) / 80)` ; :math:`g_{\mathrm{K}}=9 \mathrm{mS} / \mathrm{cm}^{2}`, and
  :math:`E_{\mathrm{K}}=-90 \mathrm{mV}`.


  Parameters
  ----------
  size: sequence of int, int
    The size of the neuron group.
  ENa: float, JaxArray, ndarray, Initializer, callable
    The reversal potential of sodium. Default is 50 mV.
  gNa: float, JaxArray, ndarray, Initializer, callable
    The maximum conductance of sodium channel. Default is 120 msiemens.
  EK: float, JaxArray, ndarray, Initializer, callable
    The reversal potential of potassium. Default is -77 mV.
  gK: float, JaxArray, ndarray, Initializer, callable
    The maximum conductance of potassium channel. Default is 36 msiemens.
  EL: float, JaxArray, ndarray, Initializer, callable
    The reversal potential of learky channel. Default is -54.387 mV.
  gL: float, JaxArray, ndarray, Initializer, callable
    The conductance of learky channel. Default is 0.03 msiemens.
  V_th: float, JaxArray, ndarray, Initializer, callable
    The threshold of the membrane spike. Default is 20 mV.
  C: float, JaxArray, ndarray, Initializer, callable
    The membrane capacitance. Default is 1 ufarad.
  phi: float, JaxArray, ndarray, Initializer, callable
    The temperature regulator constant.
  V_initializer: JaxArray, ndarray, Initializer, callable
    The initializer of membrane potential.
  h_initializer: JaxArray, ndarray, Initializer, callable
    The initializer of h channel.
  n_initializer: JaxArray, ndarray, Initializer, callable
    The initializer of n channel.
  method: str
    The numerical integration method.
  name: str
    The group name.

  References
  ----------
  .. [9] Wang, X.J. and Buzsaki, G., (1996) Gamma oscillation by synaptic
         inhibition in a hippocampal interneuronal network model. Journal of
         neuroscience, 16(20), pp.6402-6413.

  """

  def __init__(
      self,
      size: Shape,
      ENa: Union[float, Tensor, Initializer, Callable] = 55.,
      gNa: Union[float, Tensor, Initializer, Callable] = 35.,
      EK: Union[float, Tensor, Initializer, Callable] = -90.,
      gK: Union[float, Tensor, Initializer, Callable] = 9.,
      EL: Union[float, Tensor, Initializer, Callable] = -65,
      gL: Union[float, Tensor, Initializer, Callable] = 0.1,
      V_th: Union[float, Tensor, Initializer, Callable] = 20.,
      phi: Union[float, Tensor, Initializer, Callable] = 5.0,
      C: Union[float, Tensor, Initializer, Callable] = 1.0,
      V_initializer: Union[Initializer, Callable, Tensor] = OneInit(-65.),
      h_initializer: Union[Initializer, Callable, Tensor] = OneInit(0.6),
      n_initializer: Union[Initializer, Callable, Tensor] = OneInit(0.32),
      method: str = 'exp_auto',
      keep_size: bool = False,
      name: str = None
  ):
    # initialization
    super(WangBuzsakiModel, self).__init__(size=size, name=name)

    # parameters
    self.ENa = init_param(ENa, self.num, allow_none=False)
    self.EK = init_param(EK, self.num, allow_none=False)
    self.EL = init_param(EL, self.num, allow_none=False)
    self.gNa = init_param(gNa, self.num, allow_none=False)
    self.gK = init_param(gK, self.num, allow_none=False)
    self.gL = init_param(gL, self.num, allow_none=False)
    self.C = init_param(C, self.num, allow_none=False)
    self.phi = init_param(phi, self.num, allow_none=False)
    self.V_th = init_param(V_th, self.num, allow_none=False)

    # initializers
    check_initializer(h_initializer, 'h_initializer', allow_none=False)
    check_initializer(n_initializer, 'n_initializer', allow_none=False)
    check_initializer(V_initializer, 'V_initializer', allow_none=False)
    self._h_initializer = h_initializer
    self._n_initializer = n_initializer
    self._V_initializer = V_initializer

    # variables
    self.h = bm.Variable(init_param(self._h_initializer, (self.num,)))
    self.n = bm.Variable(init_param(self._n_initializer, (self.num,)))
    self.V = bm.Variable(init_param(self._V_initializer, (self.num,)))
    self.input = bm.Variable(bm.zeros(self.num))
    self.spike = bm.Variable(bm.zeros(self.num, dtype=bool))

    # integral
    self.integral = odeint(method=method, f=self.derivative)

  def reset(self):
    self.h.value = init_param(self._h_initializer, (self.num,))
    self.n.value = init_param(self._n_initializer, (self.num,))
    self.V.value = init_param(self._V_initializer, (self.num,))
    self.input[:] = 0
    self.spike[:] = False

  def m_inf(self, V):
    alpha = -0.1 * (V + 35) / (bm.exp(-0.1 * (V + 35)) - 1)
    beta = 4. * bm.exp(-(V + 60.) / 18.)
    return alpha / (alpha + beta)

  def dh(self, h, t, V):
    alpha = 0.07 * bm.exp(-(V + 58) / 20)
    beta = 1 / (bm.exp(-0.1 * (V + 28)) + 1)
    dhdt = alpha * (1 - h) - beta * h
    return self.phi * dhdt

  def dn(self, n, t, V):
    alpha = -0.01 * (V + 34) / (bm.exp(-0.1 * (V + 34)) - 1)
    beta = 0.125 * bm.exp(-(V + 44) / 80)
    dndt = alpha * (1 - n) - beta * n
    return self.phi * dndt

  def dV(self, V, t, h, n, I_ext):
    INa = self.gNa * self.m_inf(V) ** 3 * h * (V - self.ENa)
    IK = self.gK * n ** 4 * (V - self.EK)
    IL = self.gL * (V - self.EL)
    dVdt = (- INa - IK - IL + I_ext) / self.C
    return dVdt

  @property
  def derivative(self):
    return JointEq([self.dV, self.dh, self.dn])

  def update(self, t, dt):
    V, h, n = self.integral(self.V, self.h, self.n, t, self.input, dt=dt)
    self.spike.value = bm.logical_and(self.V < self.V_th, V >= self.V_th)
    self.V.value = V
    self.h.value = h
    self.n.value = n
    self.input[:] = 0.





# -*- coding: utf-8 -*-

r"""This module provides exponential integrators for ODEs.

Exponential integrators are a large class of methods from numerical analysis is based on
the exact integration of the linear part of the initial value problem. Because the linear
part is integrated exactly, this can help to mitigate the stiffness of a differential
equation.

We consider initial value problems of the form,

.. math:: u'(t)=f(u(t)),\qquad u(t_{0})=u_{0},

which can be decomposed of

.. math:: u'(t)=Lu(t)+N(u(t)),\qquad u(t_{0})=u_{0},

where :math:`L={\frac {\partial f}{\partial u}}` (the Jacobian of f) is composed of
linear terms, and :math:`N=f(u)-Lu` is composed of the non-linear terms.

This procedure enjoys the advantage, in each step, that
:math:`{\frac {\partial N_{n}}{\partial u}}(u_{n})=0`.
This considerably simplifies the derivation of the order conditions and improves the
stability when integrating the nonlinearity :math:`N(u(t))`.

Exact integration of this problem from time 0 to a later time :math:`t` can be performed
using `matrix exponentials <https://en.wikipedia.org/wiki/Matrix_exponential>`_ to define
an integral equation for the exact solution:

.. math:: u(t)=e^{Lt}u_{0}+\int _{0}^{t}e^{L(t-\tau )}N\left(t+\tau, u\left(\tau \right)\right)\,d\tau .

This representation of the exact solution is also called as *variation-of-constant formula*.
In the case of :math:`N\equiv 0`, this formulation is the exact solution to the linear
differential equation.


**Exponential Rosenbrock methods**

Exponential Rosenbrock methods were shown to be very efficient in solving large systems
of stiff ODEs. Applying the variation-of-constants formula gives the exact solution at
time :math:`t_{n+1}` with the numerical solution :math:`u_n` as

.. math::
    u(t_{n+1})=e^{h_{n}L}u(t_{n})+\int _{0}^{h_{n}}e^{(h_{n}-\tau )L}N(t_n+\tau, u(t_{n}+\tau ))d\tau .
    :label: discrete-variation-of-constants-formula

where :math:`h_n=t_{n+1}-t_n`.

The idea now is to approximate the integral in :eq:`discrete-variation-of-constants-formula`
by some quadrature rule with nodes :math:`c_{i}` and weights :math:`b_{i}(h_{n}L)`
(:math:`1\leq i\leq s`). This yields the following class of *s-stage* explicit exponential
Rosenbrock methods:

.. math::
    \begin{align}
    U_{ni}=&e^{c_{i}h_{n}L}u_n+h_{n}\sum_{j=1}^{i-1}a_{ij}(h_{n}L)N(U_{nj}),  \\
    u_{n+1}=&e^{h_{n}L}u_n+h_{n}\sum_{i=1}^{s}b_{i}(h_{n}L)N(U_{ni})
    \end{align}

where :math:`U_{ni}\approx u(t_{n}+c_{i}h_{n})`.

The coefficients :math:`a_{ij}(z),b_{i}(z)` are usually chosen as linear combinations of
the entire functions :math:`\varphi _{k}(c_{i}z),\varphi _{k}(z)`, respectively, where

.. math::
    \begin{align}
    \varphi _{k}(z)=&\int _{0}^{1}e^{(1-\theta )z}{\frac {\theta ^{k-1}}{(k-1)!}}d\theta ,\quad k\geq 1, \\
    \varphi _{0}(z)=&e^{z},\\
    \varphi _{k+1}(z)=&{\frac {\varphi_{k}(z)-\varphi _{k}(0)}{z}},\ k\geq 0.
    \end{align}

By introducing the difference :math:`D_{ni}=N(U_{ni})-N(u_{n})`, they can be reformulated
in a more efficient way for implementation as

.. math::
    \begin{align}
    U_{ni}=&u_{n}+c_{i}h_{n}\varphi _{1}(c_{i}h_{n}L)f(u_{n})+h_{n}\sum _{j=2}^{i-1}a_{ij}(h_{n}L)D_{nj}, \\
    u_{n+1}=&u_{n}+h_{n}\varphi _{1}(h_{n}L)f(u_{n})+h_{n}\sum _{i=2}^{s}b_{i}(h_{n}L)D_{ni}.
    \end{align}

where :math:`\varphi_{1}(z)=\frac{e^z-1}{z}`.

In order to implement this scheme with adaptive step size, one can consider, for the purpose
of local error estimation, the following embedded methods

.. math:: {\bar {u}}_{n+1}=u_{n}+h_{n}\varphi _{1}(h_{n}L)f(u_{n})+h_{n}\sum _{i=2}^{s}{\bar {b}}_{i}(h_{n}L)D_{ni},

which use the same stages :math:`U_{ni}` but with weights :math:`{\bar {b}}_{i}`.

For convenience, the coefficients of the explicit exponential Rosenbrock methods together
with their embedded methods can be represented by using the so-called reduced Butcher
tableau as follows:

.. math::
    \begin{array}{c|ccccc}
    c_{2} & & & & & \\
    c_{3} & a_{32} & & & & \\
    \vdots & \vdots & & \ddots & & \\
    c_{s} & a_{s 2} & a_{s 3} & \cdots & a_{s, s-1} \\
    \hline & b_{2} & b_{3} & \cdots & b_{s-1} & b_{s} \\
    & \bar{b}_{2} & \bar{b}_{3} & \cdots & \bar{b}_{s-1} & \bar{b}_{s}
    \end{array}

.. [1] https://en.wikipedia.org/wiki/Exponential_integrator
.. [2] Hochbruck, M., & Ostermann, A. (2010). Exponential integrators. Acta Numerica, 19, 209-286.
"""

import logging

from brainpy import math as bm, errors
from brainpy.base.collector import Collector
from brainpy.integrators import constants as C, utils, joint_eq
from brainpy.integrators.ode.base import ODEIntegrator
from .generic import register_ode_integrator

logger = logging.getLogger('brainpy.integrators.ode.exponential')

__all__ = [
  'ExponentialEuler',
]


class ExponentialEuler(ODEIntegrator):
  """Exponential Euler method using automatic differentiation.

  This method uses `brainpy.math.vector_grad <../../math/generated/brainpy.math.autograd.vector_grad.html>`_
  to automatically infer the linear part of the given function. Therefore, it has minimal constraints
  on your derivative function. Arbitrary complex functions can be numerically integrated with this method.

  Examples
  --------

  Here is an example uses ``ExponentialEuler`` to implement HH neuron model.

  .. plot::
    :include-source: True

    >>> import brainpy as bp
    >>> import brainpy.math as bm
    >>>
    >>> class HH(bp.dyn.NeuGroup):
    >>>   def __init__(self, size, ENa=55., EK=-90., EL=-65, C=1.0, gNa=35., gK=9.,
    >>>                gL=0.1, V_th=20., phi=5.0, name=None):
    >>>     super(HH, self).__init__(size=size, name=name)
    >>>
    >>>     # parameters
    >>>     self.ENa = ENa
    >>>     self.EK = EK
    >>>     self.EL = EL
    >>>     self.C = C
    >>>     self.gNa = gNa
    >>>     self.gK = gK
    >>>     self.gL = gL
    >>>     self.V_th = V_th
    >>>     self.phi = phi
    >>>
    >>>     # variables
    >>>     self.V = bm.Variable(bm.ones(size) * -65.)
    >>>     self.h = bm.Variable(bm.ones(size) * 0.6)
    >>>     self.n = bm.Variable(bm.ones(size) * 0.32)
    >>>     self.spike = bm.Variable(bm.zeros(size, dtype=bool))
    >>>     self.input = bm.Variable(bm.zeros(size))
    >>>
    >>>     # functions
    >>>     self.int_h = bp.ode.ExponentialEuler(self.dh)
    >>>     self.int_n = bp.ode.ExponentialEuler(self.dn)
    >>>     self.int_V = bp.ode.ExponentialEuler(self.dV)
    >>>
    >>>   def dh(self, h, t, V):
    >>>     alpha = 0.07 * bm.exp(-(V + 58) / 20)
    >>>     beta = 1 / (bm.exp(-0.1 * (V + 28)) + 1)
    >>>     dhdt = self.phi * (alpha * (1 - h) - beta * h)
    >>>     return dhdt
    >>>
    >>>   def dn(self, n, t, V):
    >>>     alpha = -0.01 * (V + 34) / (bm.exp(-0.1 * (V + 34)) - 1)
    >>>     beta = 0.125 * bm.exp(-(V + 44) / 80)
    >>>     dndt = self.phi * (alpha * (1 - n) - beta * n)
    >>>     return dndt
    >>>
    >>>   def dV(self, V, t, h, n, Iext):
    >>>     m_alpha = -0.1 * (V + 35) / (bm.exp(-0.1 * (V + 35)) - 1)
    >>>     m_beta = 4 * bm.exp(-(V + 60) / 18)
    >>>     m = m_alpha / (m_alpha + m_beta)
    >>>     INa = self.gNa * m ** 3 * h * (V - self.ENa)
    >>>     IK = self.gK * n ** 4 * (V - self.EK)
    >>>     IL = self.gL * (V - self.EL)
    >>>     dVdt = (- INa - IK - IL + Iext) / self.C
    >>>
    >>>     return dVdt
    >>>
    >>>   def update(self, _t, _dt):
    >>>     h = self.int_h(self.h, _t, self.V, dt=_dt)
    >>>     n = self.int_n(self.n, _t, self.V, dt=_dt)
    >>>     V = self.int_V(self.V, _t,  self.h, self.n, self.input, dt=_dt)
    >>>     self.spike.value = bm.logical_and(self.V < self.V_th, V >= self.V_th)
    >>>     self.V.value = V
    >>>     self.h.value = h
    >>>     self.n.value = n
    >>>     self.input[:] = 0.
    >>>
    >>> run = bp.dyn.DSRunner(HH(1), inputs=('input', 2.), monitors=['V'], dt=0.05)
    >>> run(100)
    >>> bp.visualize.line_plot(run.mon.ts, run.mon.V, legend='V', show=True)

  The above example can also be defined with ``brainpy.JointEq``.

  .. plot::
    :include-source: True

    >>> import brainpy as bp
    >>> import brainpy.math as bm
    >>>
    >>> class HH(bp.dyn.NeuGroup):
    >>>   def __init__(self, size, ENa=55., EK=-90., EL=-65, C=1.0, gNa=35., gK=9.,
    >>>                gL=0.1, V_th=20., phi=5.0, name=None):
    >>>     super(HH, self).__init__(size=size, name=name)
    >>>
    >>>     # parameters
    >>>     self.ENa = ENa
    >>>     self.EK = EK
    >>>     self.EL = EL
    >>>     self.C = C
    >>>     self.gNa = gNa
    >>>     self.gK = gK
    >>>     self.gL = gL
    >>>     self.V_th = V_th
    >>>     self.phi = phi
    >>>
    >>>     # variables
    >>>     self.V = bm.Variable(bm.ones(size) * -65.)
    >>>     self.h = bm.Variable(bm.ones(size) * 0.6)
    >>>     self.n = bm.Variable(bm.ones(size) * 0.32)
    >>>     self.spike = bm.Variable(bm.zeros(size, dtype=bool))
    >>>     self.input = bm.Variable(bm.zeros(size))
    >>>
    >>>     # functions
    >>>     derivative = bp.JointEq([self.dh, self.dn, self.dV])
    >>>     self.integral = bp.ode.ExponentialEuler(derivative)
    >>>
    >>>   def dh(self, h, t, V):
    >>>     alpha = 0.07 * bm.exp(-(V + 58) / 20)
    >>>     beta = 1 / (bm.exp(-0.1 * (V + 28)) + 1)
    >>>     dhdt = self.phi * (alpha * (1 - h) - beta * h)
    >>>     return dhdt
    >>>
    >>>   def dn(self, n, t, V):
    >>>     alpha = -0.01 * (V + 34) / (bm.exp(-0.1 * (V + 34)) - 1)
    >>>     beta = 0.125 * bm.exp(-(V + 44) / 80)
    >>>     dndt = self.phi * (alpha * (1 - n) - beta * n)
    >>>     return dndt
    >>>
    >>>   def dV(self, V, t, h, n, Iext):
    >>>     m_alpha = -0.1 * (V + 35) / (bm.exp(-0.1 * (V + 35)) - 1)
    >>>     m_beta = 4 * bm.exp(-(V + 60) / 18)
    >>>     m = m_alpha / (m_alpha + m_beta)
    >>>     INa = self.gNa * m ** 3 * h * (V - self.ENa)
    >>>     IK = self.gK * n ** 4 * (V - self.EK)
    >>>     IL = self.gL * (V - self.EL)
    >>>     dVdt = (- INa - IK - IL + Iext) / self.C
    >>>
    >>>     return dVdt
    >>>
    >>>   def update(self, t, dt):
    >>>     h, n, V = self.integral(self.h, self.n, self.V, _t, self.input, dt=_dt)
    >>>     self.spike.value = bm.logical_and(self.V < self.V_th, V >= self.V_th)
    >>>     self.V.value = V
    >>>     self.h.value = h
    >>>     self.n.value = n
    >>>     self.input[:] = 0.
    >>>
    >>> run = bp.dyn.DSRunner(HH(1), inputs=('input', 2.), monitors=['V'], dt=0.05)
    >>> run(100)
    >>> bp.visualize.line_plot(run.mon.ts, run.mon.V, legend='V', show=True)

  Parameters
  ----------
  f : function, joint_eq.JointEq
    The derivative function.
  var_type : optional, str
    The variable type.
  dt : optional, float
    The default numerical integration step.
  name : optional, str
    The integrator name.
  dyn_vars : optional, dict, sequence of JaxArray, JaxArray
  """

  def __init__(
      self,
      f,
      var_type=None,
      dt=None,
      name=None,
      show_code=False,
      dyn_vars=None,
      state_delays=None,
      neutral_delays=None
  ):
    super(ExponentialEuler, self).__init__(f=f,
                                           var_type=var_type,
                                           dt=dt,
                                           name=name,
                                           show_code=show_code,
                                           state_delays=state_delays,
                                           neutral_delays=neutral_delays)

    if var_type == C.SYSTEM_VAR:
      raise NotImplementedError(f'{self.__class__.__name__} does not support {C.SYSTEM_VAR}, '
                                f'because the auto-differentiation ')
    self.dyn_vars = dyn_vars

    # keyword checking
    keywords = {
      C.F: 'the derivative function',
    }
    utils.check_kws(self.arg_names, keywords)

    # build the integrator
    self.code_lines = []
    self.code_scope = {}
    self.integral = self.build()

  def build(self):
    all_vars, all_pars = [], []
    integrals, arg_names = [], []
    a = self._build_integrator(self.f)
    for integral, vars, _ in a:
      integrals.append(integral)
      for var in vars:
        if var not in all_vars:
          all_vars.append(var)
    for _, vars, pars in a:
      for par in pars:
        if (par not in all_vars) and (par not in all_pars):
          all_pars.append(par)
      arg_names.append(vars + pars + ['dt'])
    all_pars.append('dt')
    all_vps = all_vars + all_pars

    def integral_func(*args, **kwargs):
      # format arguments
      params_in = Collector()
      for i, arg in enumerate(args):
        params_in[all_vps[i]] = arg
      params_in.update(kwargs)
      if 'dt' not in params_in:
        params_in['dt'] = self.dt

      # call integrals
      results = []
      for i, int_fun in enumerate(integrals):
        _key = arg_names[i][0]
        r = int_fun(params_in[_key], **{arg: params_in[arg] for arg in arg_names[i][1:] if arg in params_in})
        results.append(r)
      return results if isinstance(self.f, joint_eq.JointEq) else results[0]

    return integral_func

  def _build_integrator(self, eq):
    if isinstance(eq, joint_eq.JointEq):
      results = []
      for sub_eq in eq.eqs:
        results.extend(self._build_integrator(sub_eq))
      return results

    else:
      vars, pars, _ = utils.get_args(eq)

      # checking
      if len(vars) != 1:
        raise errors.DiffEqError(C.exp_error_msg.format(cls=self.__class__.__name__,
                                                        vars=str(vars),
                                                        eq=str(eq)))

      # gradient function
      value_and_grad = bm.vector_grad(eq, argnums=0, dyn_vars=self.dyn_vars, return_value=True)

      # integration function
      def integral(*args, **kwargs):
        assert len(args) > 0
        dt = kwargs.pop('dt', self.dt)
        linear, derivative = value_and_grad(*args, **kwargs)
        phi = bm.where(linear == 0., bm.ones_like(linear), (bm.exp(dt * linear) - 1) / (dt * linear))
        return args[0] + dt * phi * derivative

      return [(integral, vars, pars), ]


register_ode_integrator('exponential_euler', ExponentialEuler)
register_ode_integrator('exp_euler', ExponentialEuler)
register_ode_integrator('exp_euler_auto', ExponentialEuler)
register_ode_integrator('exp_auto', ExponentialEuler)



# -*- coding: utf-8 -*-


from jax.lax import cond
from jax.experimental.host_callback import id_tap

__all__ = [
  'check_error_in_jit'
]


def _make_err_func(f):
  f2 = lambda arg, transforms: f(arg)

  def err_f(x):
    id_tap(f2, x)
    return
  return err_f


def check_error_in_jit(pred, err_f, err_arg=None):
  """Check errors in a jit function.

  Parameters
  ----------
  pred: bool
    The boolean prediction.
  err_f: callable
    The error function, which raise errors.
  err_arg: any
    The arguments which passed into `err_f`.
  """
  cond(pred, _make_err_func(err_f), lambda _: None, err_arg)





# -*- coding: utf-8 -*-

from typing import Dict, Type, Union, Tuple, List
import logging
import pickle

import numpy as np

from brainpy import errors
from brainpy.base.collector import TensorCollector

logger = logging.getLogger('brainpy.base.io')

__all__ = [
  'SUPPORTED_FORMATS',
  'save_as_h5',
  'save_as_npz',
  'save_as_pkl',
  'save_as_mat',
  'load_by_h5',
  'load_by_npz',
  'load_by_pkl',
  'load_by_mat',
]

SUPPORTED_FORMATS = ['.h5', '.hdf5', '.npz', '.pkl', '.mat']


def check_dict_data(
    a_dict: Dict,
    key_type: Union[Type, Tuple[Type, ...]] = None,
    val_type: Union[Type, Tuple[Type, ...]] = None,
    name: str = None
):
  """Check the dict data."""
  name = '' if (name is None) else f'"{name}"'
  if not isinstance(a_dict, dict):
    raise ValueError(f'{name} must be a dict, while we got {type(a_dict)}')
  if key_type is not None:
    for key, value in a_dict.items():
      if not isinstance(key, key_type):
        raise ValueError(f'{name} must be a dict of ({key_type}, {val_type}), '
                         f'while we got ({type(key)}, {type(value)})')
  if val_type is not None:
    for key, value in a_dict.items():
      if not isinstance(value, val_type):
        raise ValueError(f'{name} must be a dict of ({key_type}, {val_type}), '
                         f'while we got ({type(key)}, {type(value)})')


def _check_module(module, module_name, ext):
  """Check whether the required module is installed."""
  if module is None:
    raise errors.PackageMissingError(
      '"{package}" must be installed when you want to save/load data with {ext} '
      'format. \nPlease install {package} through "pip install {package}" or '
      '"conda install {package}".'.format(package=module_name, ext=ext)
    )


def _check_missing(variables, filename):
  if len(variables):
    logger.warning(f'There are variable states missed in {filename}. '
                   f'The missed variables are: {list(variables.keys())}.')


def _check_target(target):
  from .base import Base
  if not isinstance(target, Base):
    raise TypeError(f'"target" must be instance of "{Base.__name__}", but we got {type(target)}')


not_found_msg = ('"{key}" is stored in {filename}. But we does '
                 'not find it is defined as variable in {target}.')
id_dismatch_msg = ('{key1} and {key2} is the same data in {filename}. '
                   'But we found they are different in {target}.')

DUPLICATE_KEY = 'duplicate_keys'
DUPLICATE_TARGET = 'duplicate_targets'


def _load(
    target,
    verbose: bool,
    filename: str,
    load_vars: dict,
    duplicates: Tuple[List[str], List[str]],
    remove_first_axis: bool = False
):
  from brainpy import math as bm

  # get variables
  _check_target(target)
  variables = target.vars(method='absolute', level=-1)
  all_names = list(variables.keys())

  # read data from file
  for key in load_vars.keys():
    if verbose:
      print(f'Loading {key} ...')
    if key not in variables:
      raise KeyError(not_found_msg.format(key=key, target=target.name, filename=filename))
    if remove_first_axis:
      value = load_vars[key][0]
    else:
      value = load_vars[key]
    variables[key].value = bm.asarray(value)
    all_names.remove(key)

  # check duplicate names
  duplicate_keys = duplicates[0]
  duplicate_targets = duplicates[1]
  for key1, key2 in zip(duplicate_keys, duplicate_targets):
    if key1 not in all_names:
      raise KeyError(not_found_msg.format(key=key1, target=target.name, filename=filename))
    if id(variables[key1]) != id(variables[key2]):
      raise ValueError(id_dismatch_msg.format(key1=key1, key2=target, filename=filename, target=target.name))
    all_names.remove(key1)

  # check missing names
  if len(all_names):
    logger.warning(f'There are variable states missed in {filename}. '
                   f'The missed variables are: {all_names}.')


def _unique_and_duplicate(collector: dict):
  gather = TensorCollector()
  id2name = dict()
  duplicates = ([], [])
  for k, v in collector.items():
    id_ = id(v)
    if id_ not in id2name:
      gather[k] = v
      id2name[id_] = k
    else:
      k2 = id2name[id_]
      duplicates[0].append(k)
      duplicates[1].append(k2)
  duplicates = (duplicates[0], duplicates[1])
  return gather, duplicates


def save_as_h5(filename: str, variables: dict):
  """Save variables into a HDF5 file.

  Parameters
  ----------
  filename: str
    The filename to save.
  variables: dict
    All variables to save.
  """
  if not (filename.endswith('.hdf5') or filename.endswith('.h5')):
    raise ValueError(f'Cannot save variables as a HDF5 file. We only support file with '
                     f'postfix of ".hdf5" and ".h5". But we got {filename}')

  from brainpy import math as bm
  import h5py

  # check variables
  check_dict_data(variables, name='variables')
  variables, duplicates = _unique_and_duplicate(variables)

  # save
  f = h5py.File(filename, "w")
  for key, data in variables.items():
    f[key] = bm.as_numpy(data)
  if len(duplicates[0]):
    f.create_dataset(DUPLICATE_TARGET, data='+'.join(duplicates[1]))
    f.create_dataset(DUPLICATE_KEY, data='+'.join(duplicates[0]))
  f.close()


def load_by_h5(filename: str, target, verbose: bool = False):
  """Load variables in a HDF5 file.

  Parameters
  ----------
  filename: str
    The filename to load variables.
  target: Base
    The instance of :py:class:`~.brainpy.Base`.
  verbose: bool
    Whether report the load progress.
  """
  if not (filename.endswith('.hdf5') or filename.endswith('.h5')):
    raise ValueError(f'Cannot load variables from a HDF5 file. We only support file with '
                     f'postfix of ".hdf5" and ".h5". But we got {filename}')

  # read data
  import h5py
  load_vars = dict()
  with h5py.File(filename, "r") as f:
    for key in f.keys():
      if key in [DUPLICATE_KEY, DUPLICATE_TARGET]: continue
      load_vars[key] = np.asarray(f[key])
    if DUPLICATE_KEY in f:
      duplicate_keys = np.asarray(f[DUPLICATE_KEY]).item().decode("utf-8").split('+')
      duplicate_targets = np.asarray(f[DUPLICATE_TARGET]).item().decode("utf-8").split('+')
      duplicates = (duplicate_keys, duplicate_targets)
    else:
      duplicates = ([], [])

  # assign values
  _load(target, verbose, filename, load_vars, duplicates)


def save_as_npz(filename, variables, compressed=False):
  """Save variables into a numpy file.
  
  Parameters
  ----------
  filename: str
    The filename to store.
  variables: dict
    Variables to save.
  compressed: bool
    Whether we use the compressed mode.
  """
  if not filename.endswith('.npz'):
    raise ValueError(f'Cannot save variables as a .npz file. We only support file with '
                     f'postfix of ".npz". But we got {filename}')

  from brainpy import math as bm
  check_dict_data(variables, name='variables')
  variables, duplicates = _unique_and_duplicate(variables)

  # save
  variables = {k: bm.as_numpy(v) for k, v in variables.items()}
  if len(duplicates[0]):
    variables[DUPLICATE_KEY] = np.asarray(duplicates[0])
    variables[DUPLICATE_TARGET] = np.asarray(duplicates[1])
  if compressed:
    np.savez_compressed(filename, **variables)
  else:
    np.savez(filename, **variables)


def load_by_npz(filename, target, verbose=False):
  """Load variables from a numpy file.

  Parameters
  ----------
  filename: str
    The filename to load variables.
  target: Base
    The instance of :py:class:`~.brainpy.Base`.
  verbose: bool
    Whether report the load progress.
  """
  if not filename.endswith('.npz'):
    raise ValueError(f'Cannot load variables from a .npz file. We only support file with '
                     f'postfix of ".npz". But we got {filename}')

  # load data
  load_vars = dict()
  all_data = np.load(filename)
  for key in all_data.files:
    if key in [DUPLICATE_KEY, DUPLICATE_TARGET]: continue
    load_vars[key] = all_data[key]
  if DUPLICATE_KEY in all_data:
    duplicate_keys = all_data[DUPLICATE_KEY].tolist()
    duplicate_targets = all_data[DUPLICATE_TARGET].tolist()
    duplicates = (duplicate_keys, duplicate_targets)
  else:
    duplicates = ([], [])

  # assign values
  _load(target, verbose, filename, load_vars, duplicates)


def save_as_pkl(filename, variables):
  """Save variables into a pickle file.

  Parameters
  ----------
  filename: str
    The filename to save.
  variables: dict
    All variables to save.
  """
  if not (filename.endswith('.pkl') or filename.endswith('.pickle')):
    raise ValueError(f'Cannot save variables into a pickle file. We only support file with '
                     f'postfix of ".pkl" and ".pickle". But we got {filename}')

  check_dict_data(variables, name='variables')
  variables, duplicates = _unique_and_duplicate(variables)
  import brainpy.math as bm
  targets = {k: bm.as_numpy(v) for k, v in variables.items()}
  if len(duplicates[0]) > 0:
    targets[DUPLICATE_KEY] = np.asarray(duplicates[0])
    targets[DUPLICATE_TARGET] = np.asarray(duplicates[1])
  with open(filename, 'wb') as f:
    pickle.dump(targets, f, protocol=pickle.HIGHEST_PROTOCOL)


def load_by_pkl(filename, target, verbose=False):
  """Load variables from a pickle file.

  Parameters
  ----------
  filename: str
    The filename to load variables.
  target: Base
    The instance of :py:class:`~.brainpy.Base`.
  verbose: bool
    Whether report the load progress.
  """
  if not (filename.endswith('.pkl') or filename.endswith('.pickle')):
    raise ValueError(f'Cannot load variables from a pickle file. We only support file with '
                     f'postfix of ".pkl" and ".pickle". But we got {filename}')

  # load variables
  load_vars = dict()
  with open(filename, 'rb') as f:
    all_data = pickle.load(f)
    for key, data in all_data.items():
      if key in [DUPLICATE_KEY, DUPLICATE_TARGET]: continue
      load_vars[key] = data
    if DUPLICATE_KEY in all_data:
      duplicate_keys = all_data[DUPLICATE_KEY].tolist()
      duplicate_targets = all_data[DUPLICATE_TARGET].tolist()
      duplicates = (duplicate_keys, duplicate_targets)
    else:
      duplicates = ([], [])

  # assign data
  _load(target, verbose, filename, load_vars, duplicates)


def save_as_mat(filename, variables):
  """Save variables into a HDF5 file.

  Parameters
  ----------
  filename: str
    The filename to save.
  variables: dict
    All variables to save.
  """
  if not filename.endswith('.mat'):
    raise ValueError(f'Cannot save variables into a .mat file. We only support file with '
                     f'postfix of ".mat". But we got {filename}')

  from brainpy import math as bm
  import scipy.io as sio

  check_dict_data(variables, name='variables')
  variables, duplicates = _unique_and_duplicate(variables)
  variables = {k: np.expand_dims(bm.as_numpy(v), axis=0) for k, v in variables.items()}
  if len(duplicates[0]):
    variables[DUPLICATE_KEY] = np.expand_dims(np.asarray(duplicates[0]), axis=0)
    variables[DUPLICATE_TARGET] = np.expand_dims(np.asarray(duplicates[1]), axis=0)
  sio.savemat(filename, variables)


def load_by_mat(filename, target, verbose=False):
  """Load variables from a numpy file.

  Parameters
  ----------
  filename: str
    The filename to load variables.
  target: Base
    The instance of :py:class:`~.brainpy.Base`.
  verbose: bool
    Whether report the load progress.
  """
  if not filename.endswith('.mat'):
    raise ValueError(f'Cannot load variables from a .mat file. We only support file with '
                     f'postfix of ".mat". But we got {filename}')

  import scipy.io as sio

  # load data
  load_vars = dict()
  all_data = sio.loadmat(filename)
  for key, data in all_data.items():
    if key.startswith('__'):
      continue
    if key in [DUPLICATE_KEY, DUPLICATE_TARGET]:
      continue
    load_vars[key] = data[0]
  if DUPLICATE_KEY in all_data:
    duplicate_keys = [a.strip() for a in all_data[DUPLICATE_KEY].tolist()[0]]
    duplicate_targets = [a.strip() for a in all_data[DUPLICATE_TARGET].tolist()[0]]
    duplicates = (duplicate_keys, duplicate_targets)
  else:
    duplicates = ([], [])

  # assign values
  _load(target, verbose, filename, load_vars, duplicates)



# -*- coding: utf-8 -*-

from typing import Union, Callable

import brainpy.math as bm
from brainpy.dyn.base import ConNeuGroup
from brainpy.initialize import Initializer, init_param
from brainpy.integrators import odeint, JointEq
from brainpy.types import Tensor, Shape
from .base import IonChannel

__all__ = [
  'INa',
  'INa_v2',
]


class SodiumChannel(IonChannel):
  """Base class for sodium channel."""
  master_cls = ConNeuGroup


class INa(SodiumChannel):
  r"""The sodium current model.

  The sodium current model is adopted from (Bazhenov, et, al. 2002) [1]_.
  It's dynamics is given by:

  .. math::

    \begin{aligned}
    I_{\mathrm{Na}} &= g_{\mathrm{max}} * p^3 * q \\
    \frac{dp}{dt} &= \phi ( \alpha_p (1-p) - \beta_p p) \\
    \alpha_{p} &=\frac{0.32\left(V-V_{sh}-13\right)}{1-\exp \left(-\left(V-V_{sh}-13\right) / 4\right)} \\
    \beta_{p} &=\frac{-0.28\left(V-V_{sh}-40\right)}{1-\exp \left(\left(V-V_{sh}-40\right) / 5\right)} \\
    \frac{dq}{dt} & = \phi ( \alpha_q (1-h) - \beta_q h) \\
    \alpha_q &=0.128 \exp \left(-\left(V-V_{sh}-17\right) / 18\right) \\
    \beta_q &= \frac{4}{1+\exp \left(-\left(V-V_{sh}-40\right) / 5\right)}
    \end{aligned}

  where :math:`\phi` is a temperature-dependent factor, which is given by
  :math:`\phi=3^{\frac{T-36}{10}}` (:math:`T` is the temperature in Celsius).

  **Model Examples**

  - `(Brette, et, al., 2007) COBAHH <../../examples/ei_nets/Brette_2007_COBAHH.ipynb>`_

  Parameters
  ----------
  g_max : float
    The maximal conductance density (:math:`mS/cm^2`).
  E : float
    The reversal potential (mV).
  T : float
    The temperature (Celsius, :math:`^{\circ}C`).
  V_sh : float
    The shift of the membrane potential to spike.

  References
  ----------

  .. [1] Bazhenov, Maxim, et al. "Model of thalamocortical slow-wave sleep oscillations
         and transitions to activated states." Journal of neuroscience 22.19 (2002): 8691-8704.

  """

  def __init__(
      self,
      size: Shape,
      E: Union[int, float, Tensor, Initializer, Callable] = 50.,
      g_max: Union[int, float, Tensor, Initializer, Callable] = 90.,
      T: Union[int, float, Tensor, Initializer, Callable] = 36.,
      V_sh: Union[int, float, Tensor, Initializer, Callable] = -50.,
      method: str = 'exp_auto',
      name: str = None
  ):
    super(INa, self).__init__(size, name=name)

    # parameters
    self.T = init_param(T, self.num, allow_none=False)
    self.E = init_param(E, self.num, allow_none=False)
    self.V_sh = init_param(V_sh, self.num, allow_none=False)
    self.g_max = init_param(g_max, self.num, allow_none=False)
    self.phi = 3 ** ((self.T - 36) / 10)

    # variables
    self.p = bm.Variable(bm.zeros(self.num))
    self.q = bm.Variable(bm.zeros(self.num))

    # function
    self.integral = odeint(JointEq([self.dp, self.dq]), method=method)

  def reset(self, V):
    alpha = 0.32 * (V - self.V_sh - 13.) / (1. - bm.exp(-(V - self.V_sh - 13.) / 4.))
    beta = -0.28 * (V - self.V_sh - 40.) / (1. - bm.exp((V - self.V_sh - 40.) / 5.))
    self.p.value = alpha / (alpha + beta)
    alpha = 0.128 * bm.exp(-(V - self.V_sh - 17.) / 18.)
    beta = 4. / (1. + bm.exp(-(V - self.V_sh - 40.) / 5.))
    self.q.value = alpha / (alpha + beta)

  def dp(self, p, t, V):
    alpha_p = 0.32 * (V - self.V_sh - 13.) / (1. - bm.exp(-(V - self.V_sh - 13.) / 4.))
    beta_p = -0.28 * (V - self.V_sh - 40.) / (1. - bm.exp((V - self.V_sh - 40.) / 5.))
    return self.phi * (alpha_p * (1. - p) - beta_p * p)

  def dq(self, q, t, V):
    alpha_q = 0.128 * bm.exp(-(V - self.V_sh - 17.) / 18.)
    beta_q = 4. / (1. + bm.exp(-(V - self.V_sh - 40.) / 5.))
    return self.phi * (alpha_q * (1. - q) - beta_q * q)

  def update(self, t, dt, V):
    p, q = self.integral(self.p, self.q, t, V, dt)
    self.p.value, self.q.value = p, q

  def current(self, V):
    g = self.g_max * self.p ** 3 * self.q
    return g * (self.E - V)


class INa_v2(SodiumChannel):
  def __init__(
      self,
      size: Shape,
      E: Union[int, float, Tensor, Initializer, Callable] = 50.,
      g_max: Union[int, float, Tensor, Initializer, Callable] = 120.,
      method: str = 'exp_auto',
      name: str = None
  ):
    super(INa_v2, self).__init__(size, name=name)

    # parameters
    self.E = init_param(E, self.num, allow_none=False)
    self.g_max = init_param(g_max, self.num, allow_none=False)

    # variables
    self.m = bm.Variable(bm.zeros(self.num))
    self.h = bm.Variable(bm.zeros(self.num))

    # function
    self.integral = odeint(JointEq([self.dm, self.dh]), method=method)

  def dm(self, m, t, V):
    alpha = 0.1 * (V + 40) / (1 - bm.exp(-(V + 40) / 10))
    beta = 4.0 * bm.exp(-(V + 65) / 18)
    return alpha * (1 - m) - beta * m

  def dh(self, h, t, V):
    alpha = 0.07 * bm.exp(-(V + 65) / 20.)
    beta = 1 / (1 + bm.exp(-(V + 35) / 10))
    return alpha * (1 - h) - beta * h

  def update(self, t, dt, V):
    self.m.value, self.h.value = self.integral(self.m, self.h, t, V, dt)

  def current(self, V):
    g = self.g_max * self.m ** 3 * self.h
    return g * (self.E - V)

  def reset(self, V):
    alpha = 0.1 * (V + 40) / (1 - bm.exp(-(V + 40) / 10))
    beta = 4.0 * bm.exp(-(V + 65) / 18)
    self.m.value = alpha / (alpha + beta)
    alpha = 0.07 * bm.exp(-(V + 65) / 20.)
    beta = 1 / (1 + bm.exp(-(V + 35) / 10))
    self.h.value = alpha / (alpha + beta)



# -*- coding: utf-8 -*-


from . import K_channels, Na_channels, leaky_channels
from . import base, Ca_channels, Ih_channels

__all__ = []
__all__ += base.__all__
__all__ += Ca_channels.__all__
__all__ += Ih_channels.__all__
__all__ += Ih_channels.__all__
__all__ += K_channels.__all__
__all__ += leaky_channels.__all__
__all__ += Na_channels.__all__


from .base import *
from .Ca_channels import *
from .Ih_channels import *
from .K_channels import *
from .Na_channels import *
from .leaky_channels import *



# -*- coding: utf-8 -*-

from typing import Union, Callable

import brainpy.math as bm
from brainpy.dyn.base import ConNeuGroup
from brainpy.initialize import Initializer, init_param
from brainpy.integrators import odeint
from brainpy.types import Shape, Tensor
from .base import IonChannel

__all__ = [
  'PotassiumChannel',
  'IK_DR',
  'IK2',
]


class PotassiumChannel(IonChannel):
  """Base class for potassium channel."""

  '''The type of the master object.'''
  master_cls = ConNeuGroup


class IK_DR(PotassiumChannel):
  r"""The delayed rectifier potassium channel current.

  The potassium current model is adopted from (Bazhenov, et, al. 2002) [1]_.
  It's dynamics is given by:

  .. math::

      \begin{aligned}
      I_{\mathrm{K}} &= g_{\mathrm{max}} * p^4 \\
      \frac{dp}{dt} &= \phi * (\alpha_p (1-p) - \beta_p p) \\
      \alpha_{p} &=\frac{0.032\left(V-V_{sh}-15\right)}{1-\exp \left(-\left(V-V_{sh}-15\right) / 5\right)} \\
      \beta_p &= 0.5 \exp \left(-\left(V-V_{sh}-10\right) / 40\right)
      \end{aligned}

  where :math:`\phi` is a temperature-dependent factor, which is given by
  :math:`\phi=3^{\frac{T-36}{10}}` (:math:`T` is the temperature in Celsius).


  Parameters
  ----------
  size: int, sequence of int
    The object size.
  g_max : float, JaxArray, ndarray, Initializer, Callable
    The maximal conductance density (:math:`mS/cm^2`).
  E : float, JaxArray, ndarray, Initializer, Callable
    The reversal potential (mV).
  T : float, JaxArray, ndarray, Initializer, Callable
    The temperature (Celsius, :math:`^{\circ}C`).
  V_sh : float, JaxArray, ndarray, Initializer, Callable
    The shift of the membrane potential to spike.
  method: str
    The numerical integration method.
  name: str
    The object name.

  References
  ----------
  .. [1] Bazhenov, Maxim, et al. "Model of thalamocortical slow-wave sleep oscillations
         and transitions to activated states." Journal of neuroscience 22.19 (2002): 8691-8704.

  """

  def __init__(
      self,
      size: Shape,
      E: Union[float, Tensor, Initializer, Callable] = -90.,
      g_max: Union[float, Tensor, Initializer, Callable] = 10.,
      T: Union[float, Tensor, Initializer, Callable] = 36.,
      T_base: Union[float, Tensor, Initializer, Callable] = 3.,
      V_sh: Union[float, Tensor, Initializer, Callable] = -50.,
      method: str = 'exp_auto',
      name: str = None
  ):
    super(IK_DR, self).__init__(size, name=name)

    # parameters
    self.T = init_param(T, self.num, allow_none=False)
    self.T_base = init_param(T_base, self.num, allow_none=False)
    self.E = init_param(E, self.num, allow_none=False)
    self.g_max = init_param(g_max, self.num, allow_none=False)
    self.V_sh = init_param(V_sh, self.num, allow_none=False)
    self.phi = self.T_base ** ((self.T - 36) / 10)

    # variables
    self.p = bm.Variable(bm.zeros(self.num))

    # function
    self.integral = odeint(self.derivative, method=method)

  def derivative(self, p, t, V):
    alpha = 0.032 * (V - self.V_sh - 15.) / (1. - bm.exp(-(V - self.V_sh - 15.) / 5.))
    beta = 0.5 * bm.exp(-(V - self.V_sh - 10.) / 40.)
    return self.phi * (alpha * (1. - p) - beta * p)

  def update(self, t, dt, V):
    self.p.value = self.integral(self.p, t, V, dt=dt)

  def current(self, V):
    return self.g_max * self.p ** 4 * (self.E - V)

  def reset(self, V):
    alpha = 0.032 * (V - self.V_sh - 15.) / (1. - bm.exp(-(V - self.V_sh - 15.) / 5.))
    beta = 0.5 * bm.exp(-(V - self.V_sh - 10.) / 40.)
    self.p.value = alpha / (alpha + beta)


class IK2(PotassiumChannel):
  def __init__(
      self,
      size: Shape,
      E: Union[float, Tensor, Initializer, Callable] = -90.,
      g_max: Union[float, Tensor, Initializer, Callable] = 10.,
      method='exp_auto',
      name=None
  ):
    super(IK2, self).__init__(size, name=name)

    # parameters
    self.E = init_param(E, self.num, allow_none=False)
    self.g_max = init_param(g_max, self.num, allow_none=False)

    # variables
    self.n = bm.Variable(bm.zeros(self.num))

    # function
    self.integral = odeint(self.derivative, method=method)

  def derivative(self, n, t, V):
    alpha = 0.01 * (V + 55) / (1 - bm.exp(-(V + 55) / 10))
    beta = 0.125 * bm.exp(-(V + 65) / 80)
    return alpha * (1 - n) - beta * n

  def update(self, t, dt, V):
    self.n.value = self.integral(self.n, t, V, dt)

  def current(self, V):
    return self.g_max * self.n ** 4 * (self.E - V)

  def reset(self, V):
    alpha = 0.01 * (V + 55) / (1 - bm.exp(-(V + 55) / 10))
    beta = 0.125 * bm.exp(-(V + 65) / 80)
    self.n.value = alpha / (alpha + beta)
